# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from math import sqrt
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.autograd as autograd
from torch.autograd import Variable
import domainbed
import domainbed
from torch.optim.optimizer import Optimizer, required

import copy
import numpy as np
from collections import defaultdict, OrderedDict
try:
    from backpack import backpack, extend
    from backpack.extensions import BatchGrad
except:
    backpack = None

from domainbed import networks
from domainbed.lib.misc import (
    random_pairs_of_minibatches, ParamDict, MovingAverage, l2_between_dicts
)
from domainbed.losses import proxy_align_loss

ALGORITHMS = [
    'ERM',
    'Fish',
    'IRM',
    'GroupDRO',
    'Mixup',
    'MLDG',
    'CORAL',
    'MMD',
    'DANN',
    'CDANN',
    'MTL',
    'SagNet',
    'ARM',
    'VREx',
    'RSC',
    'SD',
    'ANDMask',
    'SANDMask',
    'IGA',
    'SelfReg',
    "Fishr",
    'TRM',
    'IB_ERM',
    'IB_IRM',
    'CAD',
    'CondCAD',
    'Mean_Augmentation',
    'FedMix',
    'NaiveMix',
    'Mean_Augmentation_ARM',
    'Mean_Augmentation_GroupDRO',
    'Mean_Augentation_DANN',
    'Mean_Augmentation_DCDANN',
    'Mean_Augmentation_DCDANN_Lite',
    'Mean_Augmentation_AugCA',
    'AugCA',
    'AugCA_O',
    'Moon',
    'Mean_Augmentation_Moon',
    'FedProx_algo',
    'SCAFFOLD_algo',
    'Mean_Augmentation_AugCA_Self',
    'VHL'
]

def get_algorithm_class(algorithm_name):
    """Return the algorithm class with the given name."""
    if algorithm_name not in globals():
        raise NotImplementedError("Algorithm not found: {}".format(algorithm_name))
    return globals()[algorithm_name]


class FedProx(Optimizer):
    r"""Implements FedAvg and FedProx. Local Solver can have momentum.
    Nesterov momentum is based on the formula from
    `On the importance of initialization and momentum in deep learning`__.
    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        ratio (float): relative sample size of client
        gmf (float): global/server/slow momentum factor
        mu (float): parameter for proximal local SGD
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)
    Example:
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.1, momentum=0.9)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()
    __ http://www.cs.toronto.edu/%7Ehinton/absps/momentum.pdf
    .. note::
        The implementation of SGD with Momentum/Nesterov subtly differs from
        Sutskever et. al. and implementations in some other frameworks.
        Considering the specific case of Momentum, the update can be written as
        .. math::
                  v = \rho * v + g \\
                  p = p - lr * v
        where p, g, v and :math:`\rho` denote the parameters, gradient,
        velocity, and momentum respectively.
        This is in contrast to Sutskever et. al. and
        other frameworks which employ an update of the form
        .. math::
             v = \rho * v + lr * g \\
             p = p - v
        The Nesterov version is analogously modified.
    """

    def __init__(self, params, lr=required, momentum=0, dampening=0,
                 weight_decay=0, nesterov=False, variance=0, mu=0, reduction=None):
        
        self.itr = 0
        self.a_sum = 0
        self.mu = mu


        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = dict(lr=lr, momentum=momentum, dampening=dampening,
                        weight_decay=weight_decay, nesterov=nesterov, variance=variance)
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")
        super(FedProx, self).__init__(params, defaults)


    def __setstate__(self, state):
        super(FedProx, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)

    def step(self, closure=None):
        self.mu = 0.1
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']
            
            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = p.grad.data

                if weight_decay != 0:
                    d_p.add_(weight_decay, p.data)
                
                param_state = self.state[p]
                if 'old_init' not in param_state:
                    param_state['old_init'] = torch.clone(p.data).detach()

                if momentum != 0:
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.clone(d_p).detach()
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(1 - dampening, d_p)
                    if nesterov:
                        d_p = d_p.add(momentum, buf)
                    else:
                        d_p = buf

                # apply proximal update
                d_p.add_(self.mu, p.data - param_state['old_init'])
                p.data.add_(-group['lr'], d_p)

        return loss

class SCAFFOLD_OPT(Optimizer):

    def __init__(self, params, lr=required, momentum=0, dampening=0,
                 weight_decay=0, nesterov=False, variance=0, mu=0):
        
        self.itr = 0
        self.a_sum = 0
        self.mu = mu


        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = dict(lr=lr, momentum=momentum, dampening=dampening,
                        weight_decay=weight_decay, nesterov=nesterov, variance=variance)
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")
        super(SCAFFOLD_OPT, self).__init__(params, defaults)


    def __setstate__(self, state):
        super(SCAFFOLD_OPT, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)

    def step(self, reduction, closure=None):
        self.mu = 0.1
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']
            
            for p, r in zip(group['params'], reduction):
                if p.grad is None:
                    continue
                d_p = p.grad.data

                if weight_decay != 0:
                    d_p.add_(weight_decay, p.data)
                
                param_state = self.state[p]
                if 'old_init' not in param_state:
                    param_state['old_init'] = torch.clone(p.data).detach()

                if momentum != 0:
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.clone(d_p).detach()
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(1 - dampening, d_p)
                    if nesterov:
                        d_p = d_p.add(momentum, buf)
                    else:
                        d_p = buf

                # apply proximal update
                # d_p.add_(self.mu, p.data - param_state['old_init'])
                p.data.add_(-group['lr'], d_p)
                p.data.add_(-1.0, r)
                # print(r)

        return loss

class Algorithm(torch.nn.Module):
    """
    A subclass of Algorithm implements a domain generalization algorithm.
    Subclasses should implement the following:
    - update()
    - predict()
    """
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Algorithm, self).__init__()
        self.hparams = hparams

    def update(self, minibatches, unlabeled=None):
        """
        Perform one update step, given a list of (x, y) tuples for all
        environments.

        Admits an optional list of unlabeled minibatches from the test domains,
        when task is domain_adaptation.
        """
        raise NotImplementedError

    def predict(self, x):
        raise NotImplementedError

class ERM(Algorithm):
    """
    Empirical Risk Minimization (ERM)
    """

    def get_lr(self, optimizer):
        for param_group in optimizer.param_groups:
            return param_group['lr']

    # def re_init_optimizer(self):
    #     # self.optimizer = torch.optim.Adam(
    #     #     self.network.parameters(),
    #     #     lr=self.hparams["lr"],
    #     #     weight_decay=self.hparams['weight_decay']
    #     # )
    #     self.optimizer = torch.optim.SGD(self.network.parameters(),
    #     lr=self.hparams["lr"], weight_decay=self.hparams['weight_decay'], momentum=0.9)

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(ERM, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.num_classes = num_classes
        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.classifier = networks.Classifier(
            self.featurizer.n_outputs,
            num_classes,
            self.hparams['nonlinear_classifier'])

        self.discriminator = networks.MLP(self.featurizer.n_outputs, self.featurizer.n_outputs)

        self.network = nn.Sequential(self.featurizer, self.discriminator, self.classifier)
        # self.optimizer = torch.optim.Adam(
        #     self.network.parameters(),
        #     lr=self.hparams["lr"],
        #     weight_decay=self.hparams['weight_decay']
        # )
        # self.optimizer = torch.optim.SGD(self.network.parameters(),
        # lr=self.hparams["lr"], weight_decay=self.hparams['weight_decay'], momentum=0.9)

        self.optimizer = torch.optim.SGD(self.network.parameters(),
        lr=self.hparams["lr"], weight_decay=self.hparams['weight_decay'])

        self.register_buffer('update_count', torch.tensor([0]))

        # self.optimizer = torch.optim.Adam(
        #     self.network.parameters(),
        #     lr=self.hparams["lr"],
        #     weight_decay=self.hparams['weight_decay']
        # )

    def update(self, minibatches, unlabeled=None):

        self.update_count += 1
        lr = 1.0
        # print(self.update_count)
        # if self.update_count >= 15000:
        #     lr = 0.01
        # if self.update_count >= 22500:
        #     lr = 0.001

        all_x = Variable(torch.cat([x for x,y in minibatches]))
        all_y = torch.cat([y for x,y in minibatches])
        # print(all_x.size())
        all_z = self.predict(all_x)
        loss = lr * F.cross_entropy(all_z, all_y)
        # print(self.get_lr(self.optimizer))

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': loss.item()}

    def predict(self, x):
        return self.network(x)

class VHL(ERM):

    def get_new_all_x(self, minibatches):
    
        new_minibatches = []
        for (xi, yi), (xj, yj) in random_pairs_of_minibatches(minibatches):
            lam = np.random.beta(0.2, 0.2)

            x = lam * xi + (1 - lam) * xj
            y = lam * F.one_hot(yi, self.num_classes) + (1 - lam) * F.one_hot(yj, self.num_classes) 
            new_minibatches.append((x, y))
        return new_minibatches

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(VHL, self).__init__(input_shape, num_classes, num_domains, hparams)
        self.proxy_align_loss = proxy_align_loss(
                    inter_domain_mapping=False,
                    inter_domain_class_match=True,
                    noise_feat_detach=True,
                    noise_contrastive=True,
                    inter_domain_mapping_matrix=None,
                    inter_domain_weight=0.0,
                    inter_class_weight=1.0,
                    noise_supcon_weight=0.1,
                    noise_label_shift=num_classes,
                    device=0)

    def update(self, minibatches, unlabeled=None):

        device = minibatches[0][0].device
        # new_minibatches = self.get_new_all_x(minibatches)

        VHL_alpha = 1.0
        self.update_count += 1
        # new_all_x = torch.cat([x for x,y in new_minibatches]).to(device)
        # new_all_y = torch.cat([y for x,y in new_minibatches]).to(device)
        all_x = torch.cat([x for x,y in minibatches])
        all_y = torch.cat([y for x,y in minibatches])
        all_unlabeled_x = unlabeled[0].to(device)
        all_unlabeled_y = unlabeled[1].to(device)

        # new_all_z = self.predict(new_all_x)
        # classification_loss = - torch.mean(torch.sum(F.log_softmax(new_all_z, 1) * new_all_y, 1))


        all_z = self.predict(all_x)
        classification_loss = F.cross_entropy(all_z, all_y)

        all_unlabeled_z = self.predict(all_unlabeled_x)
        aug_classification_loss = F.cross_entropy(all_unlabeled_z, all_unlabeled_y)

        loss = classification_loss + VHL_alpha * aug_classification_loss

        loss_feat_align, align_domain_loss_value, align_cls_loss_value, noise_cls_loss_value = self.proxy_align_loss(torch.cat([all_z, all_unlabeled_z]), torch.cat([all_y, all_unlabeled_y], dim=0), len(all_z))

        loss = loss + loss_feat_align
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': classification_loss.item()}



class SCAFFOLD_algo(ERM):

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(ERM, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.num_classes = num_classes
        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.classifier = networks.Classifier(
            self.featurizer.n_outputs,
            num_classes,
            self.hparams['nonlinear_classifier'])

        self.network = nn.Sequential(self.featurizer, self.classifier)
        # self.optimizer = torch.optim.Adam(
        #     self.network.parameters(),
        #     lr=self.hparams["lr"],
        #     weight_decay=self.hparams['weight_decay']
        # )
        self.optimizer = SCAFFOLD_OPT(self.network.parameters(),
        lr=self.hparams["lr"], weight_decay=self.hparams['weight_decay'], momentum=0.9)

        self.register_buffer('update_count', torch.tensor([0]))

    def get_new_all_x(self, minibatches):
        
        new_minibatches = []
        for (xi, yi), (xj, yj) in random_pairs_of_minibatches(minibatches):
            lam = np.random.beta(0.2, 0.2)

            x = lam * xi + (1 - lam) * xj
            y = lam * F.one_hot(yi, self.num_classes) + (1 - lam) * F.one_hot(yj, self.num_classes) 
            new_minibatches.append((x, y))
        return new_minibatches

    def update(self, minibatches, unlabeled=None, reduction=None):
    
        device = minibatches[0][0].device
        minibatches = self.get_new_all_x(minibatches)
        all_x = Variable(torch.cat([x for x,y in minibatches]))
        all_y = torch.cat([y for x,y in minibatches])
        # print(all_x.size())
        all_z = self.predict(all_x)
        loss = - torch.mean(torch.sum(F.log_softmax(all_z, 1) * all_y, 1))
        # print(self.get_lr(self.optimizer))

        self.optimizer.zero_grad()
        loss.backward()
        reduction = [r.to(device) for r in reduction]
        self.optimizer.step(reduction)
        reduction = [r.to('cpu') for r in reduction]

        return {'loss': loss.item()}


class FedProx_algo(ERM):

    def reset_optimizer(self):
        self.optimizer = FedProx(self.network.parameters(),
        lr=self.hparams["lr"], weight_decay=self.hparams['weight_decay'], mu = 0.1, momentum=0.9)


    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(ERM, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.num_classes = num_classes
        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.classifier = networks.Classifier(
            self.featurizer.n_outputs,
            num_classes,
            self.hparams['nonlinear_classifier'])

        # self.discriminator = networks.MLP(self.featurizer.n_outputs, self.featurizer.n_outputs)

        self.network = nn.Sequential(self.featurizer, self.classifier)

        # self.network = nn.Sequential(self.featurizer, self.classifier)
        # self.optimizer = torch.optim.Adam(
        #     self.network.parameters(),
        #     lr=self.hparams["lr"],
        #     weight_decay=self.hparams['weight_decay']
        # )
        self.optimizer = FedProx(self.network.parameters(),
        lr=self.hparams["lr"], weight_decay=self.hparams['weight_decay'], mu = 0.1, momentum=0.9)

        self.register_buffer('update_count', torch.tensor([0]))



class Moon(ERM):
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Moon, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        hparams['mlp_depth'] = 2
        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.projection_head = networks.MLP(self.featurizer.n_outputs, self.featurizer.n_outputs)
        self.classifier = networks.Classifier(
            self.projection_head.n_outputs,
            num_classes,
            self.hparams['nonlinear_classifier'])
        self.optimizer = torch.optim.SGD(list(self.featurizer.parameters()) + list(self.projection_head.parameters()) + list(self.classifier.parameters()),lr=self.hparams["lr"], weight_decay=self.hparams['weight_decay'], momentum=0.9)
        self.register_buffer('update_count', torch.tensor([0]))
        self.previous_feature = copy.deepcopy(self.featurizer)
        self.global_feature = copy.deepcopy(self.featurizer)
        self.previous_projection_head = copy.deepcopy(self.projection_head)
        self.global_projection_head = copy.deepcopy(self.projection_head)
        self.if_updated = True

    def predict(self, x):
        return self.classifier(self.projection_head(self.featurizer(x)))
        # return self.classifier(self.featurizer(x))

    def sim(self, z1, z2):
        return torch.cosine_similarity(z1, z2, dim=1)

    def get_new_all_x(self, minibatches):
        
        new_minibatches = []
        for (xi, yi), (xj, yj) in random_pairs_of_minibatches(minibatches):
            lam = np.random.beta(0.2, 0.2)

            x = lam * xi + (1 - lam) * xj
            y = lam * F.one_hot(yi, self.num_classes) + (1 - lam) * F.one_hot(yj, self.num_classes) 
            new_minibatches.append((x, y))
        return new_minibatches

    def update(self, minibatches, unlabeled=None):
        tau = 0.5
        mu = 0.01
        # minibatches = self.get_new_all_x(minibatches)
        all_x = torch.cat([x for x,y in minibatches])
        all_y = torch.cat([y for x,y in minibatches])
        all_y = F.one_hot(all_y, self.num_classes)

        if self.if_updated:
            self.global_feature = copy.deepcopy(self.featurizer)
            self.global_projection_head = copy.deepcopy(self.projection_head)
            self.if_updated = False

        z_g = self.global_projection_head(self.global_feature(all_x))
        z_p = self.previous_projection_head(self.previous_feature(all_x))
        z_c = self.projection_head(self.featurizer(all_x))

        if self.if_updated:
            l_con = 0
        else:
            l_con = -torch.log(torch.exp(self.sim(z_c, z_g) / tau) / (torch.exp(self.sim(z_c, z_g) / tau) + torch.exp(self.sim(z_c, z_p) / tau)))
            l_con = torch.sum(l_con, 0) / len(all_x)

        # z_g = self.global_feature(all_x)
        # z_p = self.previous_feature(all_x)
        # z_c = self.featurizer(all_x)

        # if self.if_updated:
        #     l_con = 0
        # else:
        #     l_con = -torch.log(torch.exp(self.sim(z_c, z_g) / tau) / (torch.exp(self.sim(z_c, z_g) / tau) + torch.exp(self.sim(z_c, z_p) / tau)))
        #     l_con = torch.sum(l_con, 0) / len(all_x)


        all_preds = self.classifier(z_c)
        l_c = - torch.mean(torch.sum(F.log_softmax(all_preds, 1) * all_y, 1))

        loss = l_c + mu * l_con

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': l_c.item(), 'penalty': l_con.item()}



class Mean_Augmentation_Moon(Moon):
    
    def update(self, minibatches, unlabeled=None):
        tau = 0.5
        mu = 1
        all_x = torch.cat([x for x,y in minibatches])
        all_y = torch.cat([y for x,y in minibatches])
        all_unlabeled = torch.cat([x for x in unlabeled])

        if self.if_updated:
            self.global_feature = copy.deepcopy(self.featurizer)
            self.if_updated = False

        z_g = self.projection_head(self.global_feature(all_unlabeled))
        z_p = self.projection_head(self.previous_feature(all_unlabeled))
        z_c = self.featurizer(all_unlabeled)

        if self.if_updated:
            l_con = 0
        else:
            l_con = -torch.log(torch.exp(self.sim(self.projection_head(z_c), z_g) / tau) / (torch.exp(self.sim(self.projection_head(z_c), z_g) / tau) + torch.exp(self.sim(self.projection_head(z_c), z_p) / tau)))
            l_con = torch.sum(l_con, 0) / len(all_x)

        l_c = F.cross_entropy(self.classifier(self.featurizer(all_x)), all_y)

        loss = l_c + mu * l_con

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': l_c.item(), 'penalty': l_con.item()}

class NaiveMix(ERM):
    def update(self, minibatches, augmentation = None):
        lambda_fedmix = 0.1
        all_x = torch.cat([x for x,y in minibatches]).requires_grad_()
        all_y = torch.cat([y for x,y in minibatches])
        all_augmentation_x = torch.cat([x for x, y in augmentation])
        all_augmentation_y = torch.zeros((len(augmentation), len(augmentation[0][1]))).to(all_x.device)
        all_x = (1 - lambda_fedmix) * all_x + lambda_fedmix * all_augmentation_x
        for i in range(len(augmentation)):
            for j in range(len(augmentation[0][1])):
                all_augmentation_y[i][j] = lambda_fedmix * augmentation[i][1][j]
            all_augmentation_y[i][all_y[i]] += 1 - lambda_fedmix


        loss = - torch.sum(F.log_softmax(self.predict(all_x), 1) * all_augmentation_y) / len(all_augmentation_y)
        # loss = F.cross_entropy(self.predict(all_x), all_augmentation_y) 

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': loss.item()}

class FedMix(ERM):


    def update(self, minibatches, augmentation = None):
        lambda_fedmix = 0.1
        all_x = torch.cat([x for x,y in minibatches]).requires_grad_()
        all_y = torch.cat([y for x,y in minibatches])
        all_augmentation_x = torch.cat([x for x, y in augmentation])
        all_augmentation_y = torch.zeros((len(augmentation), len(augmentation[0][1]))).to(all_x.device)
        for i in range(len(augmentation)):
            for j in range(len(augmentation[0][1])):
                all_augmentation_y[i][j] = augmentation[i][1][j]

        all_x = (1 - lambda_fedmix) * all_x
        objective = 0
        loss1 = (1 - lambda_fedmix) * F.cross_entropy(self.predict(all_x), all_y) 
        loss2 = -lambda_fedmix * torch.sum(F.log_softmax(self.predict(all_x), 1) * all_augmentation_y) / len(all_augmentation_y)
        grad = autograd.grad(loss1, all_x, create_graph=True)[0]
        loss3 = torch.sum(lambda_fedmix * grad * all_augmentation_x) / len(all_augmentation_y)
        objective = loss1 + loss2 + loss3

        self.optimizer.zero_grad()
        objective.backward()
        self.optimizer.step()

        return {'loss': objective.item()}







class Mean_Augmentation(ERM):

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Mean_Augmentation, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.register_buffer('update_count', torch.tensor([0]))

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                # print(m)
                m.eval()
        # print('**********************')

    def unfreeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.train()


    def update(self, minibatches, unlabeled=None):

        penalty_weight = torch.log(self.update_count * 0.1 + 1.0) + 1
        penalty_weight = 1.0
        # if self.update_count >= 25000:
        #     penalty_weight = 0.01
        # if self.update_count >= 20000:
        #     penalty_weight = 0.01
        # if self.update_count >= 30000:
        #     penalty_weight = 0.1

        all_x = torch.cat([x for x,y in minibatches])
        all_y = torch.cat([y for x,y in minibatches])
        all_unlabeled = torch.cat([x for x in unlabeled])
        self.freeze_bn()
        all_unlabeled_z = self.featurizer(all_unlabeled)
        self.unfreeze_bn()


        nll = F.cross_entropy(self.predict(all_x), all_y)
        q = torch.ones((self.num_classes, 1)).to(all_x.device) / self.num_classes
        # penalty_weight = nll.data.clone().detach()
        # penalty =  torch.std(torch.mean(torch.softmax(self.predict(all_unlabeled), 1), 0))
        penalty = - torch.sum(torch.mean(F.log_softmax(self.classifier(all_unlabeled_z), 1), 0) * q)
        loss = nll + penalty_weight * penalty
        # print(penalty)

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        self.update_count += 1

        return {'loss': nll.item(), 'penalty': penalty.item()}



class Fish(Algorithm):
    """
    Implementation of Fish, as seen in Gradient Matching for Domain 
    Generalization, Shi et al. 2021.
    """

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Fish, self).__init__(input_shape, num_classes, num_domains,
                                   hparams)
        self.input_shape = input_shape
        self.num_classes = num_classes

        self.network = networks.WholeFish(input_shape, num_classes, hparams)
        self.optimizer = torch.optim.Adam(
            self.network.parameters(),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay']
        )
        self.optimizer_inner_state = None

    def create_clone(self, device):
        self.network_inner = networks.WholeFish(self.input_shape, self.num_classes, self.hparams,
                                            weights=self.network.state_dict()).to(device)
        self.optimizer_inner = torch.optim.Adam(
            self.network_inner.parameters(),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay']
        )
        if self.optimizer_inner_state is not None:
            self.optimizer_inner.load_state_dict(self.optimizer_inner_state)

    def fish(self, meta_weights, inner_weights, lr_meta):
        meta_weights = ParamDict(meta_weights)
        inner_weights = ParamDict(inner_weights)
        meta_weights += lr_meta * (inner_weights - meta_weights)
        return meta_weights

    def update(self, minibatches, unlabeled=None):
        self.create_clone(minibatches[0][0].device)

        for x, y in minibatches:
            loss = F.cross_entropy(self.network_inner(x), y)
            self.optimizer_inner.zero_grad()
            loss.backward()
            self.optimizer_inner.step()

        self.optimizer_inner_state = self.optimizer_inner.state_dict()
        meta_weights = self.fish(
            meta_weights=self.network.state_dict(),
            inner_weights=self.network_inner.state_dict(),
            lr_meta=self.hparams["meta_lr"]
        )
        self.network.reset_weights(meta_weights)

        return {'loss': loss.item()}

    def predict(self, x):
        return self.network(x)


class ARM(ERM):
    """ Adaptive Risk Minimization (ARM) """
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        original_input_shape = input_shape
        input_shape = (1 + original_input_shape[0],) + original_input_shape[1:]
        super(ARM, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.context_net = networks.ContextNet(original_input_shape)
        self.support_size = hparams['batch_size']

    def predict(self, x):
        batch_size, c, h, w = x.shape
        if batch_size % self.support_size == 0:
            meta_batch_size = batch_size // self.support_size
            support_size = self.support_size
        else:
            meta_batch_size, support_size = 1, batch_size
        context = self.context_net(x)
        context = context.reshape((meta_batch_size, support_size, 1, h, w))
        context = context.mean(dim=1)
        context = torch.repeat_interleave(context, repeats=support_size, dim=0)
        x = torch.cat([x, context], dim=1)
        return self.network(x)

class Mean_Augmentation_ARM(ARM):

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Mean_Augmentation_ARM, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.register_buffer('update_count', torch.tensor([0]))
        self.aug_context_net = networks.ContextNet(input_shape)

    def aug_predict(self, x):
        batch_size, c, h, w = x.shape
        if batch_size % self.support_size == 0:
            meta_batch_size = batch_size // self.support_size
            support_size = self.support_size
        else:
            meta_batch_size, support_size = 1, batch_size
        context = self.aug_context_net(x)
        context = context.reshape((meta_batch_size, support_size, 1, h, w))
        context = context.mean(dim=1)
        context = torch.repeat_interleave(context, repeats=support_size, dim=0)
        x = torch.cat([x, context], dim=1)
        return self.network(x)
    


    def update(self, minibatches, unlabeled=None):

        penalty_weight = torch.log(self.update_count * 0.1 + 1.0) + 1
        penalty_weight = 0.05
        # if self.update_count >= 7000:
        #     penalty_weight = 0.01
        # if self.update_count >= 10000:
        #     penalty_weight = 0.001

        all_x = torch.cat([x for x,y in minibatches])
        all_y = torch.cat([y for x,y in minibatches])
        all_unlabeled = torch.cat([x for x in unlabeled])
        nll = F.cross_entropy(self.predict(all_x), all_y)
        q = torch.ones((self.num_classes, 1)).to(all_x.device) / self.num_classes
        penalty = - torch.sum(torch.mean(F.log_softmax(self.aug_predict(all_unlabeled), 1), 0) * q)
        loss = nll + penalty_weight * penalty

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        self.update_count += 1

        return {'loss': nll.item(), 'penalty': penalty.item()}


class AugCA(Algorithm):

    def __init__(self, input_shape, num_classes, num_domains,
                 hparams):
        
        super(AugCA, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)

        self.num_classes = num_classes
        
        self.register_buffer('update_count', torch.tensor([0]))
        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.classifier = networks.Classifier(
            self.featurizer.n_outputs,
            num_classes,
            self.hparams['nonlinear_classifier'])

        # self.feature = nn.Sequential(self.featurizer.network.conv1, self.featurizer.network.bn1, self.featurizer.network.relu, self.featurizer.network.layer1, self.featurizer.network.layer2, self.featurizer.network.avgpool)

        # self.feature = nn.Sequential(self.featurizer.network.conv1, self.featurizer.network.bn1, self.featurizer.network.relu, self.featurizer.network.layer1, self.featurizer.network.layer2, self.featurizer.network.layer3, self.featurizer.network.avgpool)

        # self.feature = copy.deepcopy(self.featurizer)

        self.hparams['mlp_width'] = 256
        self.hparams['mlp_width'] = self.featurizer.n_outputs_feature * 2
        self.discriminator = networks.MLP(self.featurizer.n_outputs_feature,
            self.featurizer.n_outputs_feature, self.hparams)

        # self.discriminator = networks.Classifier(
        #     self.featurizer.n_outputs,
        #     2,
        #     self.hparams['nonlinear_classifier'])

        self.disc_opt = torch.optim.SGD(self.discriminator.parameters(),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay'], momentum=0.9)
        
        # self.disc_opt = torch.optim.Adam(self.discriminator.parameters(),
        #     lr=self.hparams["lr"],
        #     weight_decay=self.hparams['weight_decay'])

        self.gen_opt = torch.optim.SGD(
            (list(self.featurizer.parameters()) +
                list(self.classifier.parameters())),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay'], momentum=0.9)
        # self.gen_opt = torch.optim.Adam(
        #     (list(self.featurizer.parameters()) +
        #         list(self.classifier.parameters())),
        #     lr=self.hparams["lr"],
        #     weight_decay=self.hparams['weight_decay'])

        self.if_updated = True
        self.original_feature = copy.deepcopy(self.featurizer)

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                # print(m)
                m.eval()
        # print('**********************')

    def unfreeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.train()

    def update(self, minibatches, unlabeled=None):
        # mu = 10.0 * (1.0004 ** self.update_count)
        mu = 10
        repeat = 1
        tau1 = 2.0
        tau2 = 2.0
        device = minibatches[0][0].device
        self.update_count += 1
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        all_unlabeled = torch.cat([x for x in unlabeled])

        if self.if_updated:
            self.original_feature = copy.deepcopy(self.featurizer)
            self.if_updated = False


        # all_z = self.featurizer(all_x)


        all_unlabeled_z = self.featurizer(all_unlabeled)
        all_global_unlabeled_z = self.original_feature(all_unlabeled)
        all_self_z = self.featurizer(all_x)


        embedding1 = all_unlabeled_z
        embedding2 = all_global_unlabeled_z
        embedding3 = torch.cat([all_self_z for _ in range(repeat)])

        disc_loss = - torch.log(torch.exp(self.sim(embedding1, embedding2) * tau1) / (torch.exp(self.sim(embedding1, embedding2) * tau1) + torch.exp(self.sim(embedding1, embedding3) * tau2)))
        # disc_loss = -self.sim(embedding1, embedding2) * tau1 + self.sim(embedding1, embedding3) * tau2
        disc_loss = torch.sum(disc_loss) / len(embedding1)

        all_preds = self.classifier(all_self_z)
        classifier_loss = F.cross_entropy(all_preds, all_y)
        gen_loss = (classifier_loss +
                    (mu * disc_loss))
        self.disc_opt.zero_grad()
        self.gen_opt.zero_grad()
        gen_loss.backward()
        self.gen_opt.step()
        return {'loss': classifier_loss.item(), 'penalty':disc_loss.item()}

    def predict(self, x):
        return self.classifier(self.featurizer(x))

    def sim(self, x1, x2):
        return torch.cosine_similarity(x1, x2, dim=1)


class Mean_Augmentation_AugCA(AugCA):

    def sim(self, x1, x2):
        return torch.cosine_similarity(x1, x2, dim=1)

    def get_new_unlabeled(self, all_x, all_unlabeled, K=1):
        for i in range(len(all_unlabeled)):
            # index = torch.randint(0, len(all_x), (K,))
            for j in range(K):
                index = torch.randint(0, len(all_x), (1,))
                all_unlabeled[i] = all_unlabeled[i] + all_x[index]
            all_unlabeled[i] = all_unlabeled[i] / (K + 1)
        
        return all_unlabeled

    def get_unlabeled_by_self(self, all_x, all_y, all_unlabeled, K=1):
        # one_hot_all_y = F.one_hot(all_y, self.num_classes).to(all_x.device)
        one_hot_all_y = all_y
        new_q = torch.ones((len(all_y), self.num_classes)).to(all_x.device) / self.num_classes
        for i in range(len(all_unlabeled)):
            for j in range(K):
                index = torch.randint(0, len(all_x), (1,))
                # if all_y[index] == all_y[i % len(all_y)]:
                #     index = torch.randint(0, len(all_x), (1,))
                all_unlabeled[i] = all_unlabeled[i] + all_x[index]
                new_q[i] = new_q[i] + one_hot_all_y[index]
            all_unlabeled[i] = all_unlabeled[i] / (K + 1)
            new_q[i] = new_q[i] / (K + 1)

        return all_unlabeled, new_q  

    def get_new_all_x(self, minibatches):

        new_minibatches = []
        for (xi, yi), (xj, yj) in random_pairs_of_minibatches(minibatches):
            lam = np.random.beta(0.2, 0.2)

            x = lam * xi + (1 - lam) * xj
            y = lam * F.one_hot(yi, self.num_classes) + (1 - lam) * F.one_hot(yj, self.num_classes) 
            new_minibatches.append((x, y))
        return new_minibatches



    def update(self, minibatches, unlabeled=None):
        device = minibatches[0][0].device
        mu = 0.5
        lam = 0.1
        gamma = 1.0
        tau1 = 2.0
        tau2 = 2.0
        repeat = 1
        zeta = 1.5
        # feature_dim = 16 * 4 * 4
        feature_dim = 32 * 2 * 2
        feature_dim = 256
        self.update_count += 1
        lr = 1.0
        # if self.update_count >= 15000:
        #     lr = 0.01
        # if self.update_count >= 22500:
        #     lr = 0.001

        # minibatches = self.get_new_all_x(minibatches)
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        all_y = F.one_hot(all_y, self.num_classes).to(all_x.device)
        all_unlabeled = torch.cat([x for x in unlabeled]).to(device)

        q = torch.ones((len(all_y), self.num_classes)).to(all_x.device) / self.num_classes
        # all_unlabeled = self.get_new_unlabeled(all_x, all_unlabeled, 0).to(device)
        all_unlabeled, q = self.get_unlabeled_by_self(all_x, all_y, all_unlabeled, 0)

        if self.if_updated:
            self.original_feature = copy.deepcopy(self.featurizer)
            self.if_updated = False

        all_unlabeled_z = self.featurizer(all_unlabeled)
        all_global_unlabeled_z = self.original_feature(all_unlabeled)

        all_self_z = self.featurizer(all_x)



        embedding1 = self.discriminator(all_unlabeled_z)
        embedding2 = self.discriminator(all_global_unlabeled_z)
        # embedding3 = self.discriminator(torch.cat([all_self_z for _ in range(repeat + 1)]))
        embedding3 = self.discriminator(torch.cat([all_self_z for _ in range(repeat)]))
        embedding4 = embedding1 - torch.sum(q * all_y, 1).repeat((embedding3.shape[1], 1)).T * embedding3

        # mask_1_3 = torch.eq(all_y, q)

        disc_loss = torch.log(torch.exp(self.sim(embedding1, embedding2) * tau1) / (torch.exp(self.sim(embedding1, embedding2) * tau1) + torch.exp(self.sim(embedding4, embedding3) * tau2)))

        # disc_loss = torch.log(torch.exp(self.sim(embedding1, embedding2) * tau1) / (torch.exp(self.sim(embedding1, embedding2) * tau1) + torch.exp(self.sim(embedding1, embedding3) * tau2)))

        disc_loss = lr * gamma * torch.sum(disc_loss) / len(embedding1)
        # disc_loss = (zeta ** (abs(disc_loss.item()) / (1.0)) - 1) * disc_loss
        # disc_loss = ((abs(disc_loss.item()) / (1.5)) ** zeta) * disc_loss

        self.disc_opt.zero_grad()
        disc_loss.backward()
        self.disc_opt.step()

        # all_z = self.featurizer(all_x)
        # _ = self.original_feature(all_x)


        all_self_z = self.featurizer(all_x)
        all_unlabeled_z = self.featurizer(all_unlabeled)
        all_global_unlabeled_z = self.original_feature(all_unlabeled)


        embedding1 = self.discriminator(all_unlabeled_z)
        embedding2 = self.discriminator(all_global_unlabeled_z)
        embedding3 = self.discriminator(torch.cat([all_self_z for _ in range(repeat)]))
        embedding4 = embedding1 - torch.sum(q * all_y, 1).repeat((embedding3.shape[1], 1)).T * embedding3


        disc_loss = - torch.log(torch.exp(self.sim(embedding1, embedding2) * tau1) / (torch.exp(self.sim(embedding1, embedding2) * tau1) + torch.exp(self.sim(embedding4, embedding3) * tau2)))

        # disc_loss = - torch.log(torch.exp(self.sim(embedding1, embedding2) * tau1) / (torch.exp(self.sim(embedding1, embedding2) * tau1) + torch.exp(self.sim(embedding1, embedding3) * tau2)))
        disc_loss = torch.sum(disc_loss) / len(embedding1)

        all_preds = self.classifier(all_self_z)
        classifier_loss = - torch.mean(torch.sum(F.log_softmax(all_preds, 1) * all_y, 1))

        # new_unlabeled_z = all_unlabeled_z.clone().detach()
        # new_unlabeled_z = self.featurizer(all_unlabeled)

        # q = torch.ones((self.num_classes, 1)).to(all_x.device) / self.num_classes
        # aug_penalty = - torch.sum(torch.mean(F.log_softmax(self.classifier(all_unlabeled_z), 1), 0) * q)
        aug_penalty = - torch.mean(torch.sum(torch.mul(F.log_softmax(self.classifier(all_unlabeled_z), 1), q), 1))

        gen_loss = lr * ( classifier_loss +
                    (mu * disc_loss) + lam * aug_penalty)
        # gen_loss = lr * (
        #             (mu * disc_loss) + lam * aug_penalty)
        self.disc_opt.zero_grad()
        self.gen_opt.zero_grad()
        gen_loss.backward()
        self.gen_opt.step()
        return {'loss': classifier_loss.item(), 'penalty':disc_loss.item()}


class Mean_Augmentation_AugCA_Self(AugCA):
    
    def sim(self, x1, x2):
        return torch.cosine_similarity(x1, x2, dim=1)

    def get_new_unlabeled(self, all_x, all_unlabeled, K=1):
        for i in range(len(all_unlabeled)):
            index = torch.randint(0, len(all_x), (K,))
            for j in index:
                all_unlabeled[i] = all_unlabeled[i] + all_x[j]
            all_unlabeled[i] = all_unlabeled[i] / (K + 1)
        
        return all_unlabeled


    def update(self, minibatches, unlabeled=None):
        device = minibatches[0][0].device
        mu = 0.5
        lam = 0.0
        tau1 = 2.0
        tau2 = 0.0
        repeat = 10
        self.update_count += 1
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        all_unlabeled = torch.cat([x for x in unlabeled])
        # all_unlabeled = self.get_new_unlabeled(all_x, all_unlabeled)
        # all_unlabeled = self.get_new_unlabeled(all_x, all_unlabeled, 1).to(device)


        if self.if_updated:
            self.original_feature = copy.deepcopy(self.feature).to(device)
            self.if_updated = False

        # self.freeze_bn()
        all_unlabeled_z = self.featurizer(all_unlabeled)
        all_global_unlabeled_z = self.original_feature(all_unlabeled)
        # all_unlabeled_z = self.featurizer(torch.cat([all_unlabeled, all_x]))
        # all_global_unlabeled_z = self.original_feature(torch.cat([all_unlabeled, all_x]))

        all_self_z = self.featurizer(all_x)
        # self.unfreeze_bn()



        embedding1 = self.discriminator(all_unlabeled_z)
        embedding2 = self.discriminator(all_global_unlabeled_z)
        # embedding3 = self.discriminator(torch.cat([all_self_z for _ in range(repeat + 1)]))
        embedding3 = self.discriminator(torch.cat([all_self_z for _ in range(repeat)]))

        disc_loss = torch.log(torch.exp(self.sim(embedding1, embedding2) * tau1) / (torch.exp(self.sim(embedding1, embedding2) * tau1) + torch.exp(self.sim(embedding1, embedding3) * tau2)))
        # disc_loss = self.sim(embedding1, embedding2) * tau1 - self.sim(embedding1, embedding3) * tau2
        disc_loss = torch.sum(disc_loss) / len(embedding1)

        self.disc_opt.zero_grad()
        disc_loss.backward()
        self.disc_opt.step()

        all_z = self.featurizer(all_x)
        # _ = self.original_feature(all_x)


        all_self_z = self.featurizer(all_x)
        all_unlabeled_z = self.featurizer(all_unlabeled)
        all_global_unlabeled_z = self.original_feature(all_unlabeled)
        # all_unlabeled_z = self.featurizer(torch.cat([all_unlabeled, all_x]))
        # all_global_unlabeled_z = self.original_feature(torch.cat([all_unlabeled, all_x]))


        embedding1 = self.discriminator(all_unlabeled_z)
        embedding2 = self.discriminator(all_global_unlabeled_z)
        embedding3 = self.discriminator(torch.cat([all_self_z for _ in range(repeat)]))

        disc_loss = - torch.log(torch.exp(self.sim(embedding1, embedding2) * tau1) / (torch.exp(self.sim(embedding1, embedding2) * tau1) + torch.exp(self.sim(embedding1, embedding3) * tau2)))
        # disc_loss = -self.sim(embedding1, embedding2) * tau1 + self.sim(embedding1, embedding3) * tau2
        disc_loss = torch.sum(disc_loss) / len(embedding1)

        all_preds = self.classifier(all_z)
        classifier_loss = F.cross_entropy(all_preds, all_y)

        # new_unlabeled_z = all_unlabeled_z.clone().detach()
        new_unlabeled_z = all_unlabeled_z

        q = torch.ones((self.num_classes, 1)).to(all_x.device) / self.num_classes
        aug_penalty = - torch.sum(torch.mean(F.log_softmax(self.classifier(new_unlabeled_z), 1), 0) * q)



        gen_loss = (classifier_loss +
                    (mu * disc_loss) + lam * aug_penalty)
        self.disc_opt.zero_grad()
        self.gen_opt.zero_grad()
        gen_loss.backward()
        self.gen_opt.step()
        return {'loss': classifier_loss.item(), 'penalty':disc_loss.item()}


class AugCA_O(AugCA):

    def sim(self, x1, x2):
        return torch.cosine_similarity(x1, x2, dim=1)

    def update(self, minibatches, unlabeled=None):
        device = minibatches[0][0].device
        mu = 0.5
        tau1 = 2.0
        tau2 = 1.0
        repeat = 5
        self.update_count += 1
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        all_unlabeled = torch.cat([x for x in unlabeled])

        if self.if_updated:
            self.original_feature = copy.deepcopy(self.featurizer)
            self.if_updated = False

        # self.freeze_bn()
        all_unlabeled_z = self.featurizer(all_unlabeled)
        all_global_unlabeled_z = self.original_feature(all_unlabeled)
        all_self_z = self.featurizer(all_x)
        # self.unfreeze_bn()



        embedding1 = self.discriminator(all_unlabeled_z)
        embedding2 = self.discriminator(all_global_unlabeled_z)
        embedding3 = self.discriminator(torch.cat([all_self_z for _ in range(repeat)]))
        # print(embedding1.size(), embedding2.size(), embedding3.size())

        disc_loss = torch.log(torch.exp(self.sim(embedding1, embedding2) * tau1) / (torch.exp(self.sim(embedding1, embedding2) * tau1) + torch.exp(self.sim(embedding1, embedding3) * tau2)))
        # disc_loss = self.sim(embedding1, embedding2) * tau1 - self.sim(embedding1, embedding3) * tau2
        disc_loss = torch.sum(disc_loss) / len(embedding1)

        self.disc_opt.zero_grad()
        disc_loss.backward()
        self.disc_opt.step()

        all_z = self.featurizer(all_x)
        # _ = self.original_feature(all_x)


        all_unlabeled_z = self.featurizer(all_unlabeled)
        all_global_unlabeled_z = self.original_feature(all_unlabeled)
        all_self_z = self.featurizer(all_x)


        embedding1 = self.discriminator(all_unlabeled_z)
        embedding2 = self.discriminator(all_global_unlabeled_z)
        embedding3 = self.discriminator(torch.cat([all_self_z for _ in range(repeat)]))

        disc_loss = - torch.log(torch.exp(self.sim(embedding1, embedding2) * tau1) / (torch.exp(self.sim(embedding1, embedding2) * tau1) + torch.exp(self.sim(embedding1, embedding3) * tau2)))
        # disc_loss = -self.sim(embedding1, embedding2) * tau1 + self.sim(embedding1, embedding3) * tau2
        disc_loss = torch.sum(disc_loss) / len(embedding1)

        all_preds = self.classifier(all_z)
        classifier_loss = F.cross_entropy(all_preds, all_y)
        gen_loss = (classifier_loss +
                    (mu * disc_loss))
        self.disc_opt.zero_grad()
        self.gen_opt.zero_grad()
        gen_loss.backward()
        self.gen_opt.step()
        return {'loss': classifier_loss.item(), 'penalty':disc_loss.item()}

        





class AbstractDANN(Algorithm):
    """Domain-Adversarial Neural Networks (abstract class)"""

    def __init__(self, input_shape, num_classes, num_domains,
                 hparams, conditional, class_balance):

        super(AbstractDANN, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)

        self.register_buffer('update_count', torch.tensor([0]))
        self.conditional = conditional
        self.class_balance = class_balance
        self.num_classes = num_classes

        # Algorithms
        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.classifier = networks.Classifier(
            self.featurizer.n_outputs,
            num_classes,
            self.hparams['nonlinear_classifier'])
        self.discriminator = networks.MLP(self.featurizer.n_outputs,
            num_domains, self.hparams)
        self.class_embeddings = nn.Embedding(num_classes,
            self.featurizer.n_outputs)

        # Optimizers
        self.disc_opt = torch.optim.SGD(
            (list(self.discriminator.parameters()) +
                list(self.class_embeddings.parameters())),
            lr=self.hparams["lr_d"],
            weight_decay=self.hparams['weight_decay_d'])

        self.gen_opt = torch.optim.SGD(
            (list(self.featurizer.parameters()) +
                list(self.classifier.parameters())),
            lr=self.hparams["lr_g"],
            weight_decay=self.hparams['weight_decay_g'])
        self.disc_label = 0

    def set_disc_label(self, i):
        self.disc_label = i

    def get_new_all_x(self, minibatches):
        
        new_minibatches = []
        for (xi, yi), (xj, yj) in random_pairs_of_minibatches(minibatches):
            lam = np.random.beta(0.2, 0.2)

            x = lam * xi + (1 - lam) * xj
            y = lam * F.one_hot(yi, self.num_classes) + (1 - lam) * F.one_hot(yj, self.num_classes) 
            new_minibatches.append((x, y))
        return new_minibatches

    def update(self, minibatches, unlabeled=None):
        # print(self.disc_label)
        minibatches = self.get_new_all_x(minibatches)
        device = minibatches[0][0].device
        self.update_count += 1
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        all_z = self.featurizer(all_x)
        if self.conditional:
            disc_input = all_z + self.class_embeddings(all_y)
        else:
            disc_input = all_z
        disc_out = self.discriminator(disc_input)
        disc_labels = torch.cat([
            torch.full((x.shape[0], ), self.disc_label, dtype=torch.int64, device=device)
            for i, (x, y) in enumerate(minibatches)
        ])

        if self.class_balance:
            y_counts = F.one_hot(all_y).sum(dim=0)
            weights = 1. / (y_counts[all_y] * y_counts.shape[0]).float()
            disc_loss = F.cross_entropy(disc_out, disc_labels, reduction='none')
            disc_loss = (weights * disc_loss).sum()
        else:
            disc_loss = F.cross_entropy(disc_out, disc_labels)

        disc_softmax = F.softmax(disc_out, dim=1)
        input_grad = autograd.grad(disc_softmax[:, disc_labels].sum(),
            [disc_input], create_graph=True)[0]
        grad_penalty = (input_grad**2).sum(dim=1).mean(dim=0)
        disc_loss += self.hparams['grad_penalty'] * grad_penalty

        d_steps_per_g = self.hparams['d_steps_per_g_step']

        # self.disc_opt.zero_grad()
        # disc_loss.backward()
        # self.disc_opt.step()

        # all_preds = self.classifier(all_z)
        # classifier_loss = F.cross_entropy(all_preds, all_y)
        # gen_loss = (classifier_loss +
        #             (self.hparams['lambda'] * -disc_loss))
        # self.disc_opt.zero_grad()
        # self.gen_opt.zero_grad()
        # gen_loss.backward()
        # self.gen_opt.step()
        # return {'loss': gen_loss.item()}

        if (self.update_count.item() % (1+d_steps_per_g) < d_steps_per_g):

            self.disc_opt.zero_grad()
            disc_loss.backward()
            self.disc_opt.step()
            return {'loss': disc_loss.item()}
        else:
            all_preds = self.classifier(all_z)
            classifier_loss = - torch.mean(torch.sum(F.log_softmax(all_preds, 1) * all_y, 1))
            gen_loss = (classifier_loss +
                        (self.hparams['lambda'] * -disc_loss))
            self.disc_opt.zero_grad()
            self.gen_opt.zero_grad()
            gen_loss.backward()
            self.gen_opt.step()
            return {'loss': gen_loss.item()}

    def predict(self, x):
        return self.classifier(self.featurizer(x))


class Mean_Augmentation_DANN(AbstractDANN):

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Mean_Augmentation_DANN, self).__init__(input_shape, num_classes, num_domains,
            hparams, conditional=False, class_balance=False)
        self.num_classes = num_classes

    def update(self, minibatches, unlabeled=None):
        # print(self.disc_label)
        # penalty_weight = 0.01
        # if self.update_count >= 50000:
        #     penalty_weight = 0.001
        device = minibatches[0][0].device
        self.update_count += 1
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        all_unlabeled = torch.cat([x for x in unlabeled])

        all_z = self.featurizer(all_x)
        # all_unlabeled_z = self.featurizer(all_unlabeled).clone().detach()
        all_unlabeled_z = self.featurizer(all_unlabeled)
        disc_input = all_unlabeled_z
        disc_out = self.discriminator(disc_input)
        disc_labels = torch.tensor([self.disc_label for x in all_unlabeled]).to(device)

        disc_loss = F.cross_entropy(disc_out, disc_labels)

        disc_softmax = F.softmax(disc_out, dim=1)
        input_grad = autograd.grad(disc_softmax[:, disc_labels].sum(),
            [disc_input], create_graph=True)[0]
        grad_penalty = (input_grad**2).sum(dim=1).mean(dim=0)
        disc_loss += self.hparams['grad_penalty'] * grad_penalty

        d_steps_per_g = self.hparams['d_steps_per_g_step']

        if (self.update_count.item() % (1+d_steps_per_g) < d_steps_per_g):

            self.disc_opt.zero_grad()
            disc_loss.backward()
            self.disc_opt.step()
            return {'loss': disc_loss.item()}
        else:
            all_preds = self.classifier(all_z)
            classifier_loss = F.cross_entropy(all_preds, all_y)
            gen_loss = (classifier_loss +
                        (self.hparams['lambda'] * -disc_loss))
            self.disc_opt.zero_grad()
            self.gen_opt.zero_grad()
            gen_loss.backward()
            self.gen_opt.step()
            return {'loss': classifier_loss.item(), 'penalty': disc_loss.item()}



class Mean_Augmentation_DCDANN(AbstractDANN):
    
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Mean_Augmentation_DCDANN, self).__init__(input_shape, num_classes, num_domains,
            hparams, conditional=False, class_balance=False)
        self.num_classes = num_classes
        self.num_domains = num_domains

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                # print(m)
                m.eval()
        # print('**********************')

    def unfreeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.train()

    def update(self, minibatches, unlabeled=None):
        # print(self.disc_label)
        # penalty_weight = 0.01
        # if self.update_count >= 50000:
        #     penalty_weight = 0.001
        device = minibatches[0][0].device
        self.update_count += 1
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        all_unlabeled = torch.cat([x for x in unlabeled])

        all_z = self.featurizer(all_x)
        # all_unlabeled_z = self.featurizer(all_unlabeled).clone().detach()
        self.freeze_bn()
        all_unlabeled_z = self.featurizer(all_unlabeled)
        self.unfreeze_bn()


        all_concat_z = torch.cat([all_unlabeled_z, all_z])
        disc_input = all_concat_z
        disc_out = self.discriminator(disc_input)
        disc_labels = torch.tensor([self.disc_label for x in disc_input]).to(device)

        disc_loss = F.cross_entropy(disc_out, disc_labels)

        self.disc_opt.zero_grad()
        disc_loss.backward()
        self.disc_opt.step()

        all_z = self.featurizer(all_x)
        # all_unlabeled_z = self.featurizer(all_unlabeled).clone().detach()
        self.freeze_bn()
        all_unlabeled_z = self.featurizer(all_unlabeled)
        self.unfreeze_bn()

        disc_input = all_unlabeled_z
        disc_out = self.discriminator(disc_input)
        disc_labels = torch.tensor([self.disc_label for x in disc_input]).to(device)
        # disc_labels = torch.nn.functional.one_hot(disc_labels, self.num_domains)

        disc_loss = F.cross_entropy(disc_out, disc_labels)

        all_preds = self.classifier(all_z)
        classifier_loss = F.cross_entropy(all_preds, all_y)
        gen_loss = (classifier_loss +
                    (1.0 * self.hparams['lambda'] * -disc_loss))
        self.disc_opt.zero_grad()
        self.gen_opt.zero_grad()
        gen_loss.backward()
        self.gen_opt.step()
        return {'loss': classifier_loss.item(), 'penalty':disc_loss.item()}


class Mean_Augmentation_DCDANN_Lite(AbstractDANN):
    
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Mean_Augmentation_DCDANN_Lite, self).__init__(input_shape, num_classes, 2,
            hparams, conditional=False, class_balance=False)
        self.num_classes = num_classes
        # self.original_feature = copy.deepcopy(self.featurizer)
        self.if_updated = True

    def update(self, minibatches, unlabeled=None):
        # print(self.disc_label)
        # penalty_weight = 0.01
        # if self.update_count >= 50000:
        #     penalty_weight = 0.001
        device = minibatches[0][0].device
        self.update_count += 1
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        all_unlabeled = torch.cat([x for x in unlabeled])

        all_z = self.featurizer(all_x)
        # all_unlabeled_z = self.featurizer(all_unlabeled).clone().detach()
        all_unlabeled_z = self.featurizer(all_unlabeled)

        if self.if_updated:
            self.disc_opt = torch.optim.SGD(
                (list(self.discriminator.parameters()) +
                    list(self.class_embeddings.parameters())),
                lr=self.hparams["lr_d"],
                weight_decay=self.hparams['weight_decay_d'])
            self.original_feature = copy.deepcopy(self.featurizer)
            self.if_updated = False

        d_steps_per_g = self.hparams['d_steps_per_g_step']

        if (self.update_count.item() % (1+d_steps_per_g) < d_steps_per_g):

            all_concat_z = torch.cat([all_z, self.original_feature(all_unlabeled)])
            disc_input = all_concat_z
            disc_out = self.discriminator(disc_input)
            disc_labels = torch.tensor([0 if i < len(all_z) else 1 for i in range(len(all_concat_z))]).to(device)

            disc_loss = F.cross_entropy(disc_out, disc_labels)

            disc_softmax = F.softmax(disc_out, dim=1)
            input_grad = autograd.grad(disc_softmax[:, disc_labels].sum(),
                [disc_input], create_graph=True)[0]
            grad_penalty = (input_grad**2).sum(dim=1).mean(dim=0)
            disc_loss += self.hparams['grad_penalty'] * grad_penalty

            self.disc_opt.zero_grad()
            disc_loss.backward()
            self.disc_opt.step()
            return {'loss': disc_loss.item()}
        else:
            disc_input = all_unlabeled_z
            disc_out = self.discriminator(disc_input)
            disc_labels = torch.tensor([0 for x in disc_input]).to(device)

            disc_loss = F.cross_entropy(disc_out, disc_labels)

            disc_softmax = F.softmax(disc_out, dim=1)
            input_grad = autograd.grad(disc_softmax[:, disc_labels].sum(),
                [disc_input], create_graph=True)[0]
            grad_penalty = (input_grad**2).sum(dim=1).mean(dim=0)
            disc_loss += self.hparams['grad_penalty'] * grad_penalty

            all_preds = self.classifier(all_z)
            classifier_loss = F.cross_entropy(all_preds, all_y)
            gen_loss = (classifier_loss +
                        (self.hparams['lambda'] * -disc_loss))
            self.disc_opt.zero_grad()
            self.gen_opt.zero_grad()
            gen_loss.backward()
            self.gen_opt.step()
            return {'loss': gen_loss.item()}


class DANN(AbstractDANN):
    """Unconditional DANN"""
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(DANN, self).__init__(input_shape, num_classes, num_domains,
            hparams, conditional=False, class_balance=False)


class CDANN(AbstractDANN):
    """Conditional DANN"""
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(CDANN, self).__init__(input_shape, num_classes, num_domains,
            hparams, conditional=True, class_balance=True)


class IRM(ERM):
    """Invariant Risk Minimization"""

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(IRM, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.register_buffer('update_count', torch.tensor([0]))

    @staticmethod
    def _irm_penalty(logits, y):
        device = logits[0][0].device
        scale = torch.tensor(1.).to(device).requires_grad_()
        loss_1 = F.cross_entropy(logits[::2] * scale, y[::2])
        loss_2 = F.cross_entropy(logits[1::2] * scale, y[1::2])
        grad_1 = autograd.grad(loss_1, [scale], create_graph=True)[0]
        grad_2 = autograd.grad(loss_2, [scale], create_graph=True)[0]
        result = torch.sum(grad_1 * grad_2)
        return result

    def update(self, minibatches, unlabeled=None):
        device = "cuda" if minibatches[0][0].is_cuda else "cpu"
        penalty_weight = (self.hparams['irm_lambda'] if self.update_count
                          >= self.hparams['irm_penalty_anneal_iters'] else
                          1.0)
        nll = 0.
        penalty = 0.

        all_x = torch.cat([x for x,y in minibatches])
        all_logits = self.network(all_x)
        all_logits_idx = 0
        for i, (x, y) in enumerate(minibatches):
            logits = all_logits[all_logits_idx:all_logits_idx + x.shape[0]]
            all_logits_idx += x.shape[0]
            nll += F.cross_entropy(logits, y)
            penalty += self._irm_penalty(logits, y)
        nll /= len(minibatches)
        penalty /= len(minibatches)
        # print(nll, penalty)
        loss = nll + (penalty_weight * penalty)

        # if self.update_count == self.hparams['irm_penalty_anneal_iters']:
        #     # Reset Adam, because it doesn't like the sharp jump in gradient
        #     # magnitudes that happens at this step.
        #     self.optimizer = torch.optim.SGD(
        #         self.network.parameters(),
        #         lr=self.hparams["lr"],
        #         weight_decay=self.hparams['weight_decay'])

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        self.update_count += 1
        return {'loss': loss.item(), 'nll': nll.item(),
            'penalty': penalty.item()}


class VREx(ERM):
    """V-REx algorithm from http://arxiv.org/abs/2003.00688"""
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(VREx, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.register_buffer('update_count', torch.tensor([0]))

    def update(self, minibatches, unlabeled=None):
        if self.update_count >= self.hparams["vrex_penalty_anneal_iters"]:
            penalty_weight = self.hparams["vrex_lambda"]
        else:
            penalty_weight = 1.0

        nll = 0.

        all_x = torch.cat([x for x, y in minibatches])
        all_logits = self.network(all_x)
        all_logits_idx = 0
        losses = torch.zeros(len(minibatches))
        for i, (x, y) in enumerate(minibatches):
            logits = all_logits[all_logits_idx:all_logits_idx + x.shape[0]]
            all_logits_idx += x.shape[0]
            nll = F.cross_entropy(logits, y)
            losses[i] = nll

        mean = losses.mean()
        penalty = ((losses - mean) ** 2).mean()
        loss = mean + penalty_weight * penalty

        if self.update_count == self.hparams['vrex_penalty_anneal_iters']:
            # Reset Adam (like IRM), because it doesn't like the sharp jump in
            # gradient magnitudes that happens at this step.
            self.optimizer = torch.optim.Adam(
                self.network.parameters(),
                lr=self.hparams["lr"],
                weight_decay=self.hparams['weight_decay'])

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        self.update_count += 1
        return {'loss': loss.item(), 'nll': nll.item(),
                'penalty': penalty.item()}


class Mixup(ERM):
    """
    Mixup of minibatches from different domains
    https://arxiv.org/pdf/2001.00677.pdf
    https://arxiv.org/pdf/1912.01805.pdf
    """
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Mixup, self).__init__(input_shape, num_classes, num_domains,
                                    hparams)

    def update(self, minibatches, unlabeled=None):
        objective = 0

        for (xi, yi), (xj, yj) in random_pairs_of_minibatches(minibatches):
            lam = np.random.beta(self.hparams["mixup_alpha"],
                                 self.hparams["mixup_alpha"])

            x = lam * xi + (1 - lam) * xj
            predictions = self.predict(x)

            objective += lam * F.cross_entropy(predictions, yi)
            objective += (1 - lam) * F.cross_entropy(predictions, yj)

        objective /= len(minibatches)

        self.optimizer.zero_grad()
        objective.backward()
        self.optimizer.step()

        return {'loss': objective.item()}


class GroupDRO(ERM):
    """
    Robust ERM minimizes the error at the worst minibatch
    Algorithm 1 from [https://arxiv.org/pdf/1911.08731.pdf]
    """
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(GroupDRO, self).__init__(input_shape, num_classes, num_domains,
                                        hparams)
        # self.register_buffer("q", torch.Tensor())
        self.q = None

    def get_new_all_x(self, minibatches):
        
        new_minibatches = []
        for (xi, yi), (xj, yj) in random_pairs_of_minibatches(minibatches):
            lam = np.random.beta(0.2, 0.2)

            x = lam * xi + (1 - lam) * xj
            y = lam * F.one_hot(yi, self.num_classes) + (1 - lam) * F.one_hot(yj, self.num_classes) 
            new_minibatches.append((x, y))
        return new_minibatches

    def update(self, minibatches, unlabeled=None):
        device = minibatches[0][0].device

        minibatches = self.get_new_all_x(minibatches)

        if not len(self.q):
            self.q = torch.ones(len(minibatches)).to(device)

        losses = torch.zeros(len(minibatches)).to(device)

        for m in range(len(minibatches)):
            x, y = minibatches[m]
            losses[m] = - torch.mean(torch.sum(F.log_softmax(self.predict(x), 1) * y, 1))
            self.q[m] *= (self.hparams["groupdro_eta"] * losses[m].data).exp()

        self.q /= self.q.sum()

        loss = torch.dot(losses, self.q)
        loss = loss

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': loss.item()}

class Mean_Augmentation_GroupDRO(ERM):
    """
    Robust ERM minimizes the error at the worst minibatch
    Algorithm 1 from [https://arxiv.org/pdf/1911.08731.pdf]
    """
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(Mean_Augmentation_GroupDRO, self).__init__(input_shape, num_classes, num_domains,
                                        hparams)
        # self.register_buffer("q", torch.Tensor())
        self.q = None

    def update(self, minibatches, unlabeled=None):
        device = minibatches[0][0].device

        penalty_weight = 0.5
        all_unlabeled = torch.cat([x for x in unlabeled])
        p1 = torch.ones((self.num_classes, 1)).to(device) / self.num_classes
        penalty = - torch.sum(torch.mean(F.log_softmax(self.predict(all_unlabeled), 1), 0) * p1)

        if not len(self.q):
            self.q = torch.ones(len(minibatches)).to(device)

        losses = torch.zeros(len(minibatches)).to(device)

        for m in range(len(minibatches)):
            x, y = minibatches[m]
            losses[m] = F.cross_entropy(self.predict(x), y)
            self.q[m] *= (self.hparams["groupdro_eta"] * losses[m].data).exp()

        self.q /= self.q.sum()

        nll = torch.dot(losses, self.q)
        loss = nll + penalty

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': nll.item()}


class MLDG(ERM):
    """
    Model-Agnostic Meta-Learning
    Algorithm 1 / Equation (3) from: https://arxiv.org/pdf/1710.03463.pdf
    Related: https://arxiv.org/pdf/1703.03400.pdf
    Related: https://arxiv.org/pdf/1910.13580.pdf
    """
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(MLDG, self).__init__(input_shape, num_classes, num_domains,
                                   hparams)

    def update(self, minibatches, unlabeled=None):
        """
        Terms being computed:
            * Li = Loss(xi, yi, params)
            * Gi = Grad(Li, params)

            * Lj = Loss(xj, yj, Optimizer(params, grad(Li, params)))
            * Gj = Grad(Lj, params)

            * params = Optimizer(params, Grad(Li + beta * Lj, params))
            *        = Optimizer(params, Gi + beta * Gj)

        That is, when calling .step(), we want grads to be Gi + beta * Gj

        For computational efficiency, we do not compute second derivatives.
        """
        num_mb = len(minibatches)
        objective = 0

        self.optimizer.zero_grad()
        for p in self.network.parameters():
            if p.grad is None:
                p.grad = torch.zeros_like(p)

        for (xi, yi), (xj, yj) in random_pairs_of_minibatches(minibatches):
            # fine tune clone-network on task "i"
            inner_net = copy.deepcopy(self.network)

            inner_opt = torch.optim.Adam(
                inner_net.parameters(),
                lr=self.hparams["lr"],
                weight_decay=self.hparams['weight_decay']
            )

            inner_obj = F.cross_entropy(inner_net(xi), yi)

            inner_opt.zero_grad()
            inner_obj.backward()
            inner_opt.step()

            # The network has now accumulated gradients Gi
            # The clone-network has now parameters P - lr * Gi
            for p_tgt, p_src in zip(self.network.parameters(),
                                    inner_net.parameters()):
                if p_src.grad is not None:
                    p_tgt.grad.data.add_(p_src.grad.data / num_mb)

            # `objective` is populated for reporting purposes
            objective += inner_obj.item()

            # this computes Gj on the clone-network
            loss_inner_j = F.cross_entropy(inner_net(xj), yj)
            grad_inner_j = autograd.grad(loss_inner_j, inner_net.parameters(),
                allow_unused=True)

            # `objective` is populated for reporting purposes
            objective += (self.hparams['mldg_beta'] * loss_inner_j).item()

            for p, g_j in zip(self.network.parameters(), grad_inner_j):
                if g_j is not None:
                    p.grad.data.add_(
                        self.hparams['mldg_beta'] * g_j.data / num_mb)

            # The network has now accumulated gradients Gi + beta * Gj
            # Repeat for all train-test splits, do .step()

        objective /= len(minibatches)

        self.optimizer.step()

        return {'loss': objective}

    # This commented "update" method back-propagates through the gradients of
    # the inner update, as suggested in the original MAML paper.  However, this
    # is twice as expensive as the uncommented "update" method, which does not
    # compute second-order derivatives, implementing the First-Order MAML
    # method (FOMAML) described in the original MAML paper.

    # def update(self, minibatches, unlabeled=None):
    #     objective = 0
    #     beta = self.hparams["beta"]
    #     inner_iterations = self.hparams["inner_iterations"]

    #     self.optimizer.zero_grad()

    #     with higher.innerloop_ctx(self.network, self.optimizer,
    #         copy_initial_weights=False) as (inner_network, inner_optimizer):

    #         for (xi, yi), (xj, yj) in random_pairs_of_minibatches(minibatches):
    #             for inner_iteration in range(inner_iterations):
    #                 li = F.cross_entropy(inner_network(xi), yi)
    #                 inner_optimizer.step(li)
    #
    #             objective += F.cross_entropy(self.network(xi), yi)
    #             objective += beta * F.cross_entropy(inner_network(xj), yj)

    #         objective /= len(minibatches)
    #         objective.backward()
    #
    #     self.optimizer.step()
    #
    #     return objective


class AbstractMMD(ERM):
    """
    Perform ERM while matching the pair-wise domain feature distributions
    using MMD (abstract class)
    """
    def __init__(self, input_shape, num_classes, num_domains, hparams, gaussian):
        super(AbstractMMD, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        if gaussian:
            self.kernel_type = "gaussian"
        else:
            self.kernel_type = "mean_cov"

    def my_cdist(self, x1, x2):
        x1_norm = x1.pow(2).sum(dim=-1, keepdim=True)
        x2_norm = x2.pow(2).sum(dim=-1, keepdim=True)
        res = torch.addmm(x2_norm.transpose(-2, -1),
                          x1,
                          x2.transpose(-2, -1), alpha=-2).add_(x1_norm)
        return res.clamp_min_(1e-30)

    def gaussian_kernel(self, x, y, gamma=[0.001, 0.01, 0.1, 1, 10, 100,
                                           1000]):
        D = self.my_cdist(x, y)
        K = torch.zeros_like(D)

        for g in gamma:
            K.add_(torch.exp(D.mul(-g)))

        return K

    def mmd(self, x, y):
        if self.kernel_type == "gaussian":
            Kxx = self.gaussian_kernel(x, x).mean()
            Kyy = self.gaussian_kernel(y, y).mean()
            Kxy = self.gaussian_kernel(x, y).mean()
            return Kxx + Kyy - 2 * Kxy
        else:
            mean_x = x.mean(0, keepdim=True)
            mean_y = y.mean(0, keepdim=True)
            cent_x = x - mean_x
            cent_y = y - mean_y
            cova_x = (cent_x.t() @ cent_x) / (len(x) - 1)
            cova_y = (cent_y.t() @ cent_y) / (len(y) - 1)

            mean_diff = (mean_x - mean_y).pow(2).mean()
            cova_diff = (cova_x - cova_y).pow(2).mean()

            return mean_diff + cova_diff

    def update(self, minibatches, unlabeled=None):
        objective = 0
        penalty = 0
        nmb = len(minibatches)

        features = [self.featurizer(xi) for xi, _ in minibatches]
        classifs = [self.classifier(fi) for fi in features]
        targets = [yi for _, yi in minibatches]

        for i in range(nmb):
            objective += F.cross_entropy(classifs[i], targets[i])
            for j in range(i + 1, nmb):
                penalty += self.mmd(features[i], features[j])

        objective /= nmb
        if nmb > 1:
            penalty /= (nmb * (nmb - 1) / 2)

        self.optimizer.zero_grad()
        (objective + (self.hparams['mmd_gamma']*penalty)).backward()
        self.optimizer.step()

        if torch.is_tensor(penalty):
            penalty = penalty.item()

        return {'loss': objective.item(), 'penalty': penalty}


class MMD(AbstractMMD):
    """
    MMD using Gaussian kernel
    """

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(MMD, self).__init__(input_shape, num_classes,
                                          num_domains, hparams, gaussian=True)


class CORAL(AbstractMMD):
    """
    MMD using mean and covariance difference
    """

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(CORAL, self).__init__(input_shape, num_classes,
                                         num_domains, hparams, gaussian=False)


class MTL(Algorithm):
    """
    A neural network version of
    Domain Generalization by Marginal Transfer Learning
    (https://arxiv.org/abs/1711.07910)
    """

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(MTL, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.classifier = networks.Classifier(
            self.featurizer.n_outputs * 2,
            num_classes,
            self.hparams['nonlinear_classifier'])
        self.optimizer = torch.optim.Adam(
            list(self.featurizer.parameters()) +\
            list(self.classifier.parameters()),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay']
        )

        self.register_buffer('embeddings',
                             torch.zeros(num_domains,
                                         self.featurizer.n_outputs))

        self.ema = self.hparams['mtl_ema']

    def update(self, minibatches, unlabeled=None):
        loss = 0
        for env, (x, y) in enumerate(minibatches):
            loss += F.cross_entropy(self.predict(x, env), y)

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': loss.item()}

    def update_embeddings_(self, features, env=None):
        return_embedding = features.mean(0)

        if env is not None:
            return_embedding = self.ema * return_embedding +\
                               (1 - self.ema) * self.embeddings[env]

            self.embeddings[env] = return_embedding.clone().detach()

        return return_embedding.view(1, -1).repeat(len(features), 1)

    def predict(self, x, env=None):
        features = self.featurizer(x)
        embedding = self.update_embeddings_(features, env).normal_()
        return self.classifier(torch.cat((features, embedding), 1))

class SagNet(Algorithm):
    """
    Style Agnostic Network
    Algorithm 1 from: https://arxiv.org/abs/1910.11645
    """

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(SagNet, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        # featurizer network
        self.network_f = networks.Featurizer(input_shape, self.hparams)
        # content network
        self.network_c = networks.Classifier(
            self.network_f.n_outputs,
            num_classes,
            self.hparams['nonlinear_classifier'])
        # style network
        self.network_s = networks.Classifier(
            self.network_f.n_outputs,
            num_classes,
            self.hparams['nonlinear_classifier'])

        # # This commented block of code implements something closer to the
        # # original paper, but is specific to ResNet and puts in disadvantage
        # # the other algorithms.
        # resnet_c = networks.Featurizer(input_shape, self.hparams)
        # resnet_s = networks.Featurizer(input_shape, self.hparams)
        # # featurizer network
        # self.network_f = torch.nn.Sequential(
        #         resnet_c.network.conv1,
        #         resnet_c.network.bn1,
        #         resnet_c.network.relu,
        #         resnet_c.network.maxpool,
        #         resnet_c.network.layer1,
        #         resnet_c.network.layer2,
        #         resnet_c.network.layer3)
        # # content network
        # self.network_c = torch.nn.Sequential(
        #         resnet_c.network.layer4,
        #         resnet_c.network.avgpool,
        #         networks.Flatten(),
        #         resnet_c.network.fc)
        # # style network
        # self.network_s = torch.nn.Sequential(
        #         resnet_s.network.layer4,
        #         resnet_s.network.avgpool,
        #         networks.Flatten(),
        #         resnet_s.network.fc)

        def opt(p):
            return torch.optim.Adam(p, lr=hparams["lr"],
                    weight_decay=hparams["weight_decay"])

        self.optimizer_f = opt(self.network_f.parameters())
        self.optimizer_c = opt(self.network_c.parameters())
        self.optimizer_s = opt(self.network_s.parameters())
        self.weight_adv = hparams["sag_w_adv"]

    def forward_c(self, x):
        # learning content network on randomized style
        return self.network_c(self.randomize(self.network_f(x), "style"))

    def forward_s(self, x):
        # learning style network on randomized content
        return self.network_s(self.randomize(self.network_f(x), "content"))

    def randomize(self, x, what="style", eps=1e-5):
        device = "cuda" if x.is_cuda else "cpu"
        sizes = x.size()
        alpha = torch.rand(sizes[0], 1).to(device)

        if len(sizes) == 4:
            x = x.view(sizes[0], sizes[1], -1)
            alpha = alpha.unsqueeze(-1)

        mean = x.mean(-1, keepdim=True)
        var = x.var(-1, keepdim=True)

        x = (x - mean) / (var + eps).sqrt()

        idx_swap = torch.randperm(sizes[0])
        if what == "style":
            mean = alpha * mean + (1 - alpha) * mean[idx_swap]
            var = alpha * var + (1 - alpha) * var[idx_swap]
        else:
            x = x[idx_swap].detach()

        x = x * (var + eps).sqrt() + mean
        return x.view(*sizes)

    def update(self, minibatches, unlabeled=None):
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])

        # learn content
        self.optimizer_f.zero_grad()
        self.optimizer_c.zero_grad()
        loss_c = F.cross_entropy(self.forward_c(all_x), all_y)
        loss_c.backward()
        self.optimizer_f.step()
        self.optimizer_c.step()

        # learn style
        self.optimizer_s.zero_grad()
        loss_s = F.cross_entropy(self.forward_s(all_x), all_y)
        loss_s.backward()
        self.optimizer_s.step()

        # learn adversary
        self.optimizer_f.zero_grad()
        loss_adv = -F.log_softmax(self.forward_s(all_x), dim=1).mean(1).mean()
        loss_adv = loss_adv * self.weight_adv
        loss_adv.backward()
        self.optimizer_f.step()

        return {'loss_c': loss_c.item(), 'loss_s': loss_s.item(),
                'loss_adv': loss_adv.item()}

    def predict(self, x):
        return self.network_c(self.network_f(x))


class RSC(ERM):
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(RSC, self).__init__(input_shape, num_classes, num_domains,
                                   hparams)
        self.drop_f = (1 - hparams['rsc_f_drop_factor']) * 100
        self.drop_b = (1 - hparams['rsc_b_drop_factor']) * 100
        self.num_classes = num_classes

    def update(self, minibatches, unlabeled=None):
        device = "cuda" if minibatches[0][0].is_cuda else "cpu"

        # inputs
        all_x = torch.cat([x for x, y in minibatches])
        # labels
        all_y = torch.cat([y for _, y in minibatches])
        # one-hot labels
        all_o = torch.nn.functional.one_hot(all_y, self.num_classes)
        # features
        all_f = self.featurizer(all_x)
        # predictions
        all_p = self.classifier(all_f)

        # Equation (1): compute gradients with respect to representation
        all_g = autograd.grad((all_p * all_o).sum(), all_f)[0]

        # Equation (2): compute top-gradient-percentile mask
        percentiles = np.percentile(all_g.cpu(), self.drop_f, axis=1)
        percentiles = torch.Tensor(percentiles)
        percentiles = percentiles.unsqueeze(1).repeat(1, all_g.size(1))
        mask_f = all_g.lt(percentiles.to(device)).float()

        # Equation (3): mute top-gradient-percentile activations
        all_f_muted = all_f * mask_f

        # Equation (4): compute muted predictions
        all_p_muted = self.classifier(all_f_muted)

        # Section 3.3: Batch Percentage
        all_s = F.softmax(all_p, dim=1)
        all_s_muted = F.softmax(all_p_muted, dim=1)
        changes = (all_s * all_o).sum(1) - (all_s_muted * all_o).sum(1)
        percentile = np.percentile(changes.detach().cpu(), self.drop_b)
        mask_b = changes.lt(percentile).float().view(-1, 1)
        mask = torch.logical_or(mask_f, mask_b).float()

        # Equations (3) and (4) again, this time mutting over examples
        all_p_muted_again = self.classifier(all_f * mask)

        # Equation (5): update
        loss = F.cross_entropy(all_p_muted_again, all_y)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': loss.item()}


class SD(ERM):
    """
    Gradient Starvation: A Learning Proclivity in Neural Networks
    Equation 25 from [https://arxiv.org/pdf/2011.09468.pdf]
    """
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(SD, self).__init__(input_shape, num_classes, num_domains,
                                        hparams)
        self.sd_reg = hparams["sd_reg"]

    def update(self, minibatches, unlabeled=None):
        all_x = torch.cat([x for x,y in minibatches])
        all_y = torch.cat([y for x,y in minibatches])
        all_p = self.predict(all_x)

        loss = F.cross_entropy(all_p, all_y)
        penalty = (all_p ** 2).mean()
        objective = loss + self.sd_reg * penalty

        self.optimizer.zero_grad()
        objective.backward()
        self.optimizer.step()

        return {'loss': loss.item(), 'penalty': penalty.item()}

class ANDMask(ERM):
    """
    Learning Explanations that are Hard to Vary [https://arxiv.org/abs/2009.00329]
    AND-Mask implementation from [https://github.com/gibipara92/learning-explanations-hard-to-vary]
    """

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(ANDMask, self).__init__(input_shape, num_classes, num_domains, hparams)

        self.tau = hparams["tau"]

    def update(self, minibatches, unlabeled=None):
        mean_loss = 0
        param_gradients = [[] for _ in self.network.parameters()]
        for i, (x, y) in enumerate(minibatches):
            logits = self.network(x)
            
            env_loss = F.cross_entropy(logits, y)
            mean_loss += env_loss.item() / len(minibatches)

            env_grads = autograd.grad(env_loss, self.network.parameters())
            for grads, env_grad in zip(param_gradients, env_grads):
                grads.append(env_grad)

        self.optimizer.zero_grad()
        self.mask_grads(self.tau, param_gradients, self.network.parameters())
        self.optimizer.step()

        return {'loss': mean_loss}

    def mask_grads(self, tau, gradients, params):

        for param, grads in zip(params, gradients):
            grads = torch.stack(grads, dim=0)
            grad_signs = torch.sign(grads)
            mask = torch.mean(grad_signs, dim=0).abs() >= self.tau
            mask = mask.to(torch.float32)
            avg_grad = torch.mean(grads, dim=0)

            mask_t = (mask.sum() / mask.numel())
            param.grad = mask * avg_grad
            param.grad *= (1. / (1e-10 + mask_t))

        return 0

class IGA(ERM):
    """
    Inter-environmental Gradient Alignment
    From https://arxiv.org/abs/2008.01883v2
    """

    def __init__(self, in_features, num_classes, num_domains, hparams):
        super(IGA, self).__init__(in_features, num_classes, num_domains, hparams)

    def update(self, minibatches, unlabeled=False):
        total_loss = 0
        grads = []
        for i, (x, y) in enumerate(minibatches):
            logits = self.network(x)

            env_loss = F.cross_entropy(logits, y)
            total_loss += env_loss

            env_grad = autograd.grad(env_loss, self.network.parameters(), 
                                        create_graph=True)

            grads.append(env_grad)
            
        mean_loss = total_loss / len(minibatches)
        mean_grad = autograd.grad(mean_loss, self.network.parameters(), 
                                        retain_graph=True)

        # compute trace penalty
        penalty_value = 0
        for grad in grads:
            for g, mean_g in zip(grad, mean_grad):
                penalty_value += (g - mean_g).pow(2).sum()

        objective = mean_loss + self.hparams['penalty'] * penalty_value

        self.optimizer.zero_grad()
        objective.backward()
        self.optimizer.step()

        return {'loss': mean_loss.item(), 'penalty': penalty_value.item()}
    
    
class SelfReg(ERM):
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(SelfReg, self).__init__(input_shape, num_classes, num_domains,
                                   hparams)
        self.num_classes = num_classes
        self.MSEloss = nn.MSELoss()
        input_feat_size = self.featurizer.n_outputs
        hidden_size = input_feat_size if input_feat_size==2048 else input_feat_size*2
        
        self.cdpl = nn.Sequential(
                            nn.Linear(input_feat_size, hidden_size),
                            nn.BatchNorm1d(hidden_size),
                            nn.ReLU(inplace=True),
                            nn.Linear(hidden_size, hidden_size),
                            nn.BatchNorm1d(hidden_size),
                            nn.ReLU(inplace=True),
                            nn.Linear(hidden_size, input_feat_size),
                            nn.BatchNorm1d(input_feat_size)
        )
        
    def update(self, minibatches, unlabeled=None):
        
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for _, y in minibatches])

        lam = np.random.beta(0.5, 0.5)
        
        batch_size = all_y.size()[0]
        
        # cluster and order features into same-class group
        with torch.no_grad():   
            sorted_y, indices = torch.sort(all_y)
            sorted_x = torch.zeros_like(all_x)
            for idx, order in enumerate(indices):
                sorted_x[idx] = all_x[order]
            intervals = []
            ex = 0
            for idx, val in enumerate(sorted_y):
                if ex==val:
                    continue
                intervals.append(idx)
                ex = val
            intervals.append(batch_size)

            all_x = sorted_x
            all_y = sorted_y
        
        feat = self.featurizer(all_x)
        proj = self.cdpl(feat)
        
        output = self.classifier(feat)

        # shuffle
        output_2 = torch.zeros_like(output)
        feat_2 = torch.zeros_like(proj)
        output_3 = torch.zeros_like(output)
        feat_3 = torch.zeros_like(proj)
        ex = 0
        for end in intervals:
            shuffle_indices = torch.randperm(end-ex)+ex
            shuffle_indices2 = torch.randperm(end-ex)+ex
            for idx in range(end-ex):
                output_2[idx+ex] = output[shuffle_indices[idx]]
                feat_2[idx+ex] = proj[shuffle_indices[idx]]
                output_3[idx+ex] = output[shuffle_indices2[idx]]
                feat_3[idx+ex] = proj[shuffle_indices2[idx]]
            ex = end
        
        # mixup 
        output_3 = lam*output_2 + (1-lam)*output_3
        feat_3 = lam*feat_2 + (1-lam)*feat_3

        # regularization
        L_ind_logit = self.MSEloss(output, output_2)
        L_hdl_logit = self.MSEloss(output, output_3)
        L_ind_feat = 0.3 * self.MSEloss(feat, feat_2)
        L_hdl_feat = 0.3 * self.MSEloss(feat, feat_3)
        
        cl_loss = F.cross_entropy(output, all_y)
        C_scale = min(cl_loss.item(), 1.)
        loss = cl_loss + C_scale*(lam*(L_ind_logit + L_ind_feat)+(1-lam)*(L_hdl_logit + L_hdl_feat))
     
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        return {'loss': loss.item()}


class SANDMask(ERM):
    """
    SAND-mask: An Enhanced Gradient Masking Strategy for the Discovery of Invariances in Domain Generalization
    <https://arxiv.org/abs/2106.02266>
    """

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(SANDMask, self).__init__(input_shape, num_classes, num_domains, hparams)

        self.tau = hparams["tau"]
        self.k = hparams["k"]
        betas = (0.9, 0.999)
        self.optimizer = torch.optim.Adam(
            self.network.parameters(),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay'],
            betas=betas
        )

        self.register_buffer('update_count', torch.tensor([0]))

    def update(self, minibatches, unlabeled=None):

        mean_loss = 0
        param_gradients = [[] for _ in self.network.parameters()]
        for i, (x, y) in enumerate(minibatches):
            logits = self.network(x)

            env_loss = F.cross_entropy(logits, y)
            mean_loss += env_loss.item() / len(minibatches)
            env_grads = autograd.grad(env_loss, self.network.parameters(), retain_graph=True)
            for grads, env_grad in zip(param_gradients, env_grads):
                grads.append(env_grad)

        self.optimizer.zero_grad()
        # gradient masking applied here
        self.mask_grads(param_gradients, self.network.parameters())
        self.optimizer.step()
        self.update_count += 1

        return {'loss': mean_loss}

    def mask_grads(self, gradients, params):
        '''
        Here a mask with continuous values in the range [0,1] is formed to control the amount of update for each
        parameter based on the agreement of gradients coming from different environments.
        '''
        device = gradients[0][0].device
        for param, grads in zip(params, gradients):
            grads = torch.stack(grads, dim=0)
            avg_grad = torch.mean(grads, dim=0)
            grad_signs = torch.sign(grads)
            gamma = torch.tensor(1.0).to(device)
            grads_var = grads.var(dim=0)
            grads_var[torch.isnan(grads_var)] = 1e-17
            lam = (gamma * grads_var).pow(-1)
            mask = torch.tanh(self.k * lam * (torch.abs(grad_signs.mean(dim=0)) - self.tau))
            mask = torch.max(mask, torch.zeros_like(mask))
            mask[torch.isnan(mask)] = 1e-17
            mask_t = (mask.sum() / mask.numel())
            param.grad = mask * avg_grad
            param.grad *= (1. / (1e-10 + mask_t))



class Fishr(Algorithm):
    "Invariant Gradients variances for Out-of-distribution Generalization"

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        assert backpack is not None, "Install backpack with: 'pip install backpack-for-pytorch==1.3.0'"
        super(Fishr, self).__init__(input_shape, num_classes, num_domains, hparams)
        self.num_domains = num_domains

        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.classifier = extend(
            networks.Classifier(
                self.featurizer.n_outputs,
                num_classes,
                self.hparams['nonlinear_classifier'],
            )
        )
        self.network = nn.Sequential(self.featurizer, self.classifier)

        self.register_buffer("update_count", torch.tensor([0]))
        self.bce_extended = extend(nn.CrossEntropyLoss(reduction='none'))
        self.ema_per_domain = [
            MovingAverage(ema=self.hparams["ema"], oneminusema_correction=True)
            for _ in range(self.num_domains)
        ]
        self._init_optimizer()

    def _init_optimizer(self):
        self.optimizer = torch.optim.Adam(
            list(self.featurizer.parameters()) + list(self.classifier.parameters()),
            lr=self.hparams["lr"],
            weight_decay=self.hparams["weight_decay"],
        )

    def update(self, minibatches, unlabeled=False):
        assert len(minibatches) == self.num_domains
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        len_minibatches = [x.shape[0] for x, y in minibatches]

        all_z = self.featurizer(all_x)
        all_logits = self.classifier(all_z)

        penalty = self.compute_fishr_penalty(all_logits, all_y, len_minibatches)
        all_nll = F.cross_entropy(all_logits, all_y)

        penalty_weight = 0
        if self.update_count >= self.hparams["penalty_anneal_iters"]:
            penalty_weight = self.hparams["lambda"]
            if self.update_count == self.hparams["penalty_anneal_iters"] != 0:
                # Reset Adam as in IRM or V-REx, because it may not like the sharp jump in
                # gradient magnitudes that happens at this step.
                self._init_optimizer()
        self.update_count += 1

        objective = all_nll + penalty_weight * penalty
        self.optimizer.zero_grad()
        objective.backward()
        self.optimizer.step()

        return {'loss': objective.item(), 'nll': all_nll.item(), 'penalty': penalty.item()}

    def compute_fishr_penalty(self, all_logits, all_y, len_minibatches):
        dict_grads = self._get_grads(all_logits, all_y)
        grads_var_per_domain = self._get_grads_var_per_domain(dict_grads, len_minibatches)
        return self._compute_distance_grads_var(grads_var_per_domain)

    def _get_grads(self, logits, y):
        self.optimizer.zero_grad()
        loss = self.bce_extended(logits, y).sum()
        with backpack(BatchGrad()):
            loss.backward(
                inputs=list(self.classifier.parameters()), retain_graph=True, create_graph=True
            )

        # compute individual grads for all samples across all domains simultaneously
        dict_grads = OrderedDict(
            [
                (name, weights.grad_batch.clone().view(weights.grad_batch.size(0), -1))
                for name, weights in self.classifier.named_parameters()
            ]
        )
        return dict_grads

    def _get_grads_var_per_domain(self, dict_grads, len_minibatches):
        # grads var per domain
        grads_var_per_domain = [{} for _ in range(self.num_domains)]
        for name, _grads in dict_grads.items():
            all_idx = 0
            for domain_id, bsize in enumerate(len_minibatches):
                env_grads = _grads[all_idx:all_idx + bsize]
                all_idx += bsize
                env_mean = env_grads.mean(dim=0, keepdim=True)
                env_grads_centered = env_grads - env_mean
                grads_var_per_domain[domain_id][name] = (env_grads_centered).pow(2).mean(dim=0)

        # moving average
        for domain_id in range(self.num_domains):
            grads_var_per_domain[domain_id] = self.ema_per_domain[domain_id].update(
                grads_var_per_domain[domain_id]
            )

        return grads_var_per_domain

    def _compute_distance_grads_var(self, grads_var_per_domain):

        # compute gradient variances averaged across domains
        grads_var = OrderedDict(
            [
                (
                    name,
                    torch.stack(
                        [
                            grads_var_per_domain[domain_id][name]
                            for domain_id in range(self.num_domains)
                        ],
                        dim=0
                    ).mean(dim=0)
                )
                for name in grads_var_per_domain[0].keys()
            ]
        )

        penalty = 0
        for domain_id in range(self.num_domains):
            penalty += l2_between_dicts(grads_var_per_domain[domain_id], grads_var)
        return penalty / self.num_domains

    def predict(self, x):
        return self.network(x)

class TRM(Algorithm):
    """
    Learning Representations that Support Robust Transfer of Predictors
    <https://arxiv.org/abs/2110.09940>
    """

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(TRM, self).__init__(input_shape, num_classes, num_domains,hparams)
        self.register_buffer('update_count', torch.tensor([0]))
        self.num_domains = num_domains
        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.classifier = nn.Linear(self.featurizer.n_outputs, num_classes).cuda()
        self.clist = [nn.Linear(self.featurizer.n_outputs, num_classes).cuda() for i in range(num_domains+1)]
        self.olist = [torch.optim.SGD(
            self.clist[i].parameters(),
            lr=1e-1,
        ) for i in range(num_domains+1)]

        self.optimizer_f = torch.optim.Adam(
            self.featurizer.parameters(),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay']
        )
        self.optimizer_c = torch.optim.Adam(
            self.classifier.parameters(),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay']
        )
        # initial weights
        self.alpha = torch.ones((num_domains, num_domains)).cuda() - torch.eye(num_domains).cuda()

    @staticmethod
    def neum(v, model, batch):
        def hvp(y, w, v):

            # First backprop
            first_grads = autograd.grad(y, w, retain_graph=True, create_graph=True, allow_unused=True)
            first_grads = torch.nn.utils.parameters_to_vector(first_grads)
            # Elementwise products
            elemwise_products = first_grads @ v
            # Second backprop
            return_grads = autograd.grad(elemwise_products, w, create_graph=True)
            return_grads = torch.nn.utils.parameters_to_vector(return_grads)
            return return_grads

        v = v.detach()
        h_estimate = v
        cnt = 0.
        model.eval()
        iter = 10
        for i in range(iter):
            model.weight.grad *= 0
            y = model(batch[0].detach())
            loss = F.cross_entropy(y, batch[1].detach())
            hv = hvp(loss, model.weight, v)
            v -= hv
            v = v.detach()
            h_estimate = v + h_estimate
            h_estimate = h_estimate.detach()
            # not converge
            if torch.max(abs(h_estimate)) > 10:
                break
            cnt += 1

        model.train()
        return h_estimate.detach()

    def update(self, minibatches):

        loss_swap = 0.0
        trm = 0.0

        if self.update_count >= self.hparams['iters']:
            # TRM
            if self.hparams['class_balanced']:
                # for stability when facing unbalanced labels across environments
                for classifier in self.clist:
                    classifier.weight.data = copy.deepcopy(self.classifier.weight.data)
            self.alpha /= self.alpha.sum(1, keepdim=True)
            
            self.featurizer.train()
            all_x = torch.cat([x for x, y in minibatches])
            all_y = torch.cat([y for x, y in minibatches])
            all_feature = self.featurizer(all_x)
            # updating original network
            loss = F.cross_entropy(self.classifier(all_feature), all_y)

            for i in range(30):
                all_logits_idx = 0
                loss_erm = 0.
                for j, (x, y) in enumerate(minibatches):
                    # j-th domain
                    feature = all_feature[all_logits_idx:all_logits_idx + x.shape[0]]
                    all_logits_idx += x.shape[0]
                    loss_erm += F.cross_entropy(self.clist[j](feature.detach()), y)
                for opt in self.olist:
                    opt.zero_grad()
                loss_erm.backward()
                for opt in self.olist:
                    opt.step()

            # collect (feature, y)
            feature_split = list()
            y_split = list()
            all_logits_idx = 0
            for i, (x, y) in enumerate(minibatches):
                feature = all_feature[all_logits_idx:all_logits_idx + x.shape[0]]
                all_logits_idx += x.shape[0]
                feature_split.append(feature)
                y_split.append(y)

            # estimate transfer risk
            for Q, (x, y) in enumerate(minibatches):
                sample_list = list(range(len(minibatches)))
                sample_list.remove(Q)

                loss_Q = F.cross_entropy(self.clist[Q](feature_split[Q]), y_split[Q])
                grad_Q = autograd.grad(loss_Q, self.clist[Q].weight, create_graph=True)
                vec_grad_Q = nn.utils.parameters_to_vector(grad_Q)

                loss_P = [F.cross_entropy(self.clist[Q](feature_split[i]), y_split[i])*(self.alpha[Q, i].data.detach())
                          if i in sample_list else 0. for i in range(len(minibatches))]
                loss_P_sum = sum(loss_P)
                grad_P = autograd.grad(loss_P_sum, self.clist[Q].weight, create_graph=True)
                vec_grad_P = nn.utils.parameters_to_vector(grad_P).detach()
                vec_grad_P = self.neum(vec_grad_P, self.clist[Q], (feature_split[Q], y_split[Q]))

                loss_swap += loss_P_sum - self.hparams['cos_lambda'] * (vec_grad_P.detach() @ vec_grad_Q)

                for i in sample_list:
                    self.alpha[Q, i] *= (self.hparams["groupdro_eta"] * loss_P[i].data).exp()

            loss_swap /= len(minibatches)
            trm /= len(minibatches)
        else:
            # ERM
            self.featurizer.train()
            all_x = torch.cat([x for x, y in minibatches])
            all_y = torch.cat([y for x, y in minibatches])
            all_feature = self.featurizer(all_x)
            loss = F.cross_entropy(self.classifier(all_feature), all_y)

        nll = loss.item()
        self.optimizer_c.zero_grad()
        self.optimizer_f.zero_grad()
        if self.update_count >= self.hparams['iters']:
            loss_swap = (loss + loss_swap)
        else:
            loss_swap = loss

        loss_swap.backward()
        self.optimizer_f.step()
        self.optimizer_c.step()

        loss_swap = loss_swap.item() - nll
        self.update_count += 1

        return {'nll': nll, 'trm_loss': loss_swap}

    def predict(self, x):
        return self.classifier(self.featurizer(x))

    def train(self):
        self.featurizer.train()

    def eval(self):
        self.featurizer.eval()

class IB_ERM(ERM):
    """Information Bottleneck based ERM on feature with conditionning"""

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(IB_ERM, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.optimizer = torch.optim.Adam(
            list(self.featurizer.parameters()) + list(self.classifier.parameters()),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay']
        )
        self.register_buffer('update_count', torch.tensor([0]))

    def update(self, minibatches, unlabeled=None):
        device = "cuda" if minibatches[0][0].is_cuda else "cpu"
        ib_penalty_weight = (self.hparams['ib_lambda'] if self.update_count
                          >= self.hparams['ib_penalty_anneal_iters'] else
                          0.0)

        nll = 0.
        ib_penalty = 0.

        all_x = torch.cat([x for x,y in minibatches])
        all_features = self.featurizer(all_x)
        all_logits = self.classifier(all_features)
        all_logits_idx = 0
        for i, (x, y) in enumerate(minibatches):
            features = all_features[all_logits_idx:all_logits_idx + x.shape[0]]
            logits = all_logits[all_logits_idx:all_logits_idx + x.shape[0]]
            all_logits_idx += x.shape[0]
            nll += F.cross_entropy(logits, y)
            ib_penalty += features.var(dim=0).mean()

        nll /= len(minibatches)
        ib_penalty /= len(minibatches)

        # Compile loss
        loss = nll 
        loss += ib_penalty_weight * ib_penalty

        if self.update_count == self.hparams['ib_penalty_anneal_iters']:
            # Reset Adam, because it doesn't like the sharp jump in gradient
            # magnitudes that happens at this step.
            self.optimizer = torch.optim.Adam(
                list(self.featurizer.parameters()) + list(self.classifier.parameters()),
                lr=self.hparams["lr"],
                weight_decay=self.hparams['weight_decay'])

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        self.update_count += 1
        return {'loss': loss.item(), 
                'nll': nll.item(),
                'IB_penalty': ib_penalty.item()}

class IB_IRM(ERM):
    """Information Bottleneck based IRM on feature with conditionning"""

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(IB_IRM, self).__init__(input_shape, num_classes, num_domains,
                                  hparams)
        self.optimizer = torch.optim.Adam(
            list(self.featurizer.parameters()) + list(self.classifier.parameters()),
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay']
        )
        self.register_buffer('update_count', torch.tensor([0]))

    @staticmethod
    def _irm_penalty(logits, y):
        device = "cuda" if logits[0][0].is_cuda else "cpu"
        scale = torch.tensor(1.).to(device).requires_grad_()
        loss_1 = F.cross_entropy(logits[::2] * scale, y[::2])
        loss_2 = F.cross_entropy(logits[1::2] * scale, y[1::2])
        grad_1 = autograd.grad(loss_1, [scale], create_graph=True)[0]
        grad_2 = autograd.grad(loss_2, [scale], create_graph=True)[0]
        result = torch.sum(grad_1 * grad_2)
        return result

    def update(self, minibatches, unlabeled=None):
        device = "cuda" if minibatches[0][0].is_cuda else "cpu"
        irm_penalty_weight = (self.hparams['irm_lambda'] if self.update_count
                          >= self.hparams['irm_penalty_anneal_iters'] else
                          1.0)
        ib_penalty_weight = (self.hparams['ib_lambda'] if self.update_count
                          >= self.hparams['ib_penalty_anneal_iters'] else
                          0.0)

        nll = 0.
        irm_penalty = 0.
        ib_penalty = 0.

        all_x = torch.cat([x for x,y in minibatches])
        all_features = self.featurizer(all_x)
        all_logits = self.classifier(all_features)
        all_logits_idx = 0
        for i, (x, y) in enumerate(minibatches):
            features = all_features[all_logits_idx:all_logits_idx + x.shape[0]]
            logits = all_logits[all_logits_idx:all_logits_idx + x.shape[0]]
            all_logits_idx += x.shape[0]
            nll += F.cross_entropy(logits, y)
            irm_penalty += self._irm_penalty(logits, y)
            ib_penalty += features.var(dim=0).mean()

        nll /= len(minibatches)
        irm_penalty /= len(minibatches)
        ib_penalty /= len(minibatches)

        # Compile loss
        loss = nll 
        loss += irm_penalty_weight * irm_penalty
        loss += ib_penalty_weight * ib_penalty

        if self.update_count == self.hparams['irm_penalty_anneal_iters'] or self.update_count == self.hparams['ib_penalty_anneal_iters']:
            # Reset Adam, because it doesn't like the sharp jump in gradient
            # magnitudes that happens at this step.
            self.optimizer = torch.optim.Adam(
                list(self.featurizer.parameters()) + list(self.classifier.parameters()),
                lr=self.hparams["lr"],
                weight_decay=self.hparams['weight_decay'])

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        self.update_count += 1
        return {'loss': loss.item(), 
                'nll': nll.item(),
                'IRM_penalty': irm_penalty.item(), 
                'IB_penalty': ib_penalty.item()}


class AbstractCAD(Algorithm):
    """Contrastive adversarial domain bottleneck (abstract class)
    from Optimal Representations for Covariate Shift <https://arxiv.org/abs/2201.00057>
    """

    def __init__(self, input_shape, num_classes, num_domains,
                 hparams, is_conditional):
        super(AbstractCAD, self).__init__(input_shape, num_classes, num_domains, hparams)

        self.featurizer = networks.Featurizer(input_shape, self.hparams)
        self.classifier = networks.Classifier(
            self.featurizer.n_outputs,
            num_classes,
            self.hparams['nonlinear_classifier'])
        params = list(self.featurizer.parameters()) + list(self.classifier.parameters())

        # parameters for domain bottleneck loss
        self.is_conditional = is_conditional  # whether to use bottleneck conditioned on the label
        self.base_temperature = 0.07
        self.temperature = hparams['temperature']
        self.is_project = hparams['is_project']  # whether apply projection head
        self.is_normalized = hparams['is_normalized'] # whether apply normalization to representation when computing loss

        # whether flip maximize log(p) (False) to minimize -log(1-p) (True) for the bottleneck loss
        # the two versions have the same optima, but we find the latter is more stable
        self.is_flipped = hparams["is_flipped"]

        if self.is_project:
            self.project = nn.Sequential(
                nn.Linear(feature_dim, feature_dim),
                nn.ReLU(inplace=True),
                nn.Linear(feature_dim, 128),
            )
            params += list(self.project.parameters())

        # Optimizers
        self.optimizer = torch.optim.Adam(
            params,
            lr=self.hparams["lr"],
            weight_decay=self.hparams['weight_decay']
        )

    def bn_loss(self, z, y, dom_labels):
        """Contrastive based domain bottleneck loss
         The implementation is based on the supervised contrastive loss (SupCon) introduced by
         P. Khosla, et al., in “Supervised Contrastive Learning“.
        Modified from  https://github.com/HobbitLong/SupContrast/blob/8d0963a7dbb1cd28accb067f5144d61f18a77588/losses.py#L11
        """
        device = z.device
        batch_size = z.shape[0]

        y = y.contiguous().view(-1, 1)
        dom_labels = dom_labels.contiguous().view(-1, 1)
        mask_y = torch.eq(y, y.T).to(device)
        mask_d = (torch.eq(dom_labels, dom_labels.T)).to(device)
        mask_drop = ~torch.eye(batch_size).bool().to(device)  # drop the "current"/"self" example
        mask_y &= mask_drop
        mask_y_n_d = mask_y & (~mask_d)  # contain the same label but from different domains
        mask_y_d = mask_y & mask_d  # contain the same label and the same domain
        mask_y, mask_drop, mask_y_n_d, mask_y_d = mask_y.float(), mask_drop.float(), mask_y_n_d.float(), mask_y_d.float()

        # compute logits
        if self.is_project:
            z = self.project(z)
        if self.is_normalized:
            z = F.normalize(z, dim=1)
        outer = z @ z.T
        logits = outer / self.temperature
        logits = logits * mask_drop
        # for numerical stability
        logits_max, _ = torch.max(logits, dim=1, keepdim=True)
        logits = logits - logits_max.detach()

        if not self.is_conditional:
            # unconditional CAD loss
            denominator = torch.logsumexp(logits + mask_drop.log(), dim=1, keepdim=True)
            log_prob = logits - denominator

            mask_valid = (mask_y.sum(1) > 0)
            log_prob = log_prob[mask_valid]
            mask_d = mask_d[mask_valid]

            if self.is_flipped:  # maximize log prob of samples from different domains
                bn_loss = - (self.temperature / self.base_temperature) * torch.logsumexp(
                    log_prob + (~mask_d).float().log(), dim=1)
            else:  # minimize log prob of samples from same domain
                bn_loss = (self.temperature / self.base_temperature) * torch.logsumexp(
                    log_prob + (mask_d).float().log(), dim=1)
        else:
            # conditional CAD loss
            if self.is_flipped:
                mask_valid = (mask_y_n_d.sum(1) > 0)
            else:
                mask_valid = (mask_y_d.sum(1) > 0)

            mask_y = mask_y[mask_valid]
            mask_y_d = mask_y_d[mask_valid]
            mask_y_n_d = mask_y_n_d[mask_valid]
            logits = logits[mask_valid]

            # compute log_prob_y with the same label
            denominator = torch.logsumexp(logits + mask_y.log(), dim=1, keepdim=True)
            log_prob_y = logits - denominator

            if self.is_flipped:  # maximize log prob of samples from different domains and with same label
                bn_loss = - (self.temperature / self.base_temperature) * torch.logsumexp(
                    log_prob_y + mask_y_n_d.log(), dim=1)
            else:  # minimize log prob of samples from same domains and with same label
                bn_loss = (self.temperature / self.base_temperature) * torch.logsumexp(
                    log_prob_y + mask_y_d.log(), dim=1)

        def finite_mean(x):
            # only 1D for now
            num_finite = (torch.isfinite(x).float()).sum()
            mean = torch.where(torch.isfinite(x), x, torch.tensor(0.0).to(x)).sum()
            if num_finite != 0:
                mean = mean / num_finite
            else:
                return torch.tensor(0.0).to(x)
            return mean

        return finite_mean(bn_loss)

    def update(self, minibatches, unlabeled=None):
        device = "cuda" if minibatches[0][0].is_cuda else "cpu"
        all_x = torch.cat([x for x, y in minibatches])
        all_y = torch.cat([y for x, y in minibatches])
        all_z = self.featurizer(all_x)
        all_d = torch.cat([
            torch.full((x.shape[0],), i, dtype=torch.int64, device=device)
            for i, (x, y) in enumerate(minibatches)
        ])

        bn_loss = self.bn_loss(all_z, all_y, all_d)
        clf_out = self.classifier(all_z)
        clf_loss = F.cross_entropy(clf_out, all_y)
        total_loss = clf_loss + self.hparams['lmbda'] * bn_loss

        self.optimizer.zero_grad()
        total_loss.backward()
        self.optimizer.step()

        return {"clf_loss": clf_loss.item(), "bn_loss": bn_loss.item(), "total_loss": total_loss.item()}

    def predict(self, x):
        return self.classifier(self.featurizer(x))


class CAD(AbstractCAD):
    """Contrastive Adversarial Domain (CAD) bottleneck

       Properties:
       - Minimize I(D;Z)
       - Require access to domain labels but not task labels
       """

    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(CAD, self).__init__(input_shape, num_classes, num_domains, hparams, is_conditional=False)


class CondCAD(AbstractCAD):
    """Conditional Contrastive Adversarial Domain (CAD) bottleneck

    Properties:
    - Minimize I(D;Z|Y)
    - Require access to both domain labels and task labels
    """
    def __init__(self, input_shape, num_classes, num_domains, hparams):
        super(CondCAD, self).__init__(input_shape, num_classes, num_domains, hparams, is_conditional=True)
