# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models
from domainbed.Resnet import ResNet_cifar

from domainbed.lib import wide_resnet
from domainbed.src import cct_7_3x1_32_c100, cct_7_3x1_32
import copy


def remove_batch_norm_from_resnet(model):
    fuse = torch.nn.utils.fusion.fuse_conv_bn_eval
    model.eval()

    model.conv1 = fuse(model.conv1, model.bn1)
    model.bn1 = Identity()

    for name, module in model.named_modules():
        if name.startswith("layer") and len(name) == 6:
            for b, bottleneck in enumerate(module):
                for name2, module2 in bottleneck.named_modules():
                    if name2.startswith("conv"):
                        bn_name = "bn" + name2[-1]
                        setattr(bottleneck, name2,
                                fuse(module2, getattr(bottleneck, bn_name)))
                        setattr(bottleneck, bn_name, Identity())
                if isinstance(bottleneck.downsample, torch.nn.Sequential):
                    bottleneck.downsample[0] = fuse(bottleneck.downsample[0],
                                                    bottleneck.downsample[1])
                    bottleneck.downsample[1] = Identity()
    model.train()
    return model


class Identity(nn.Module):
    """An identity layer"""
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


class MLP(nn.Module):
    """Just  an MLP"""
    def __init__(self, n_inputs, n_outputs):
        super(MLP, self).__init__()
        self.input = nn.Linear(n_inputs, n_inputs)
        # self.dropout = nn.Dropout(hparams['mlp_dropout'])
        self.hiddens = nn.ModuleList([
            nn.Linear(n_inputs, n_inputs)
            for _ in range(1)])
        self.output = nn.Linear(n_inputs, n_outputs)
        self.n_outputs = n_outputs

    def forward(self, x):
        x = self.input(x)
        # x = self.dropout(x)
        x = F.relu(x)
        for hidden in self.hiddens:
            x = hidden(x)
            # x = self.dropout(x)
            x = F.relu(x)
        x = self.output(x)
        return x

class ViT(torch.nn.Module):
    def __init__(self, input_shape, hparams):
        super(ViT, self).__init__()
        # self.network = cct_7_3x1_32_c100()
        self.network = cct_7_3x1_32()
        self.n_outputs = 256
        self.n_outputs_feature = 256
        
    def forward(self, x):
        """Encode x into a feature vector of size n_outputs."""
        return self.network(x)

    def train(self, mode=True):
        """
        Override the default train() to freeze the BN parameters
        """
        super().train(mode)
        # self.freeze_bn()
    



class ResNet(torch.nn.Module):
    """ResNet with the softmax chopped off and the batchnorm frozen"""
    def __init__(self, input_shape, hparams):
        super(ResNet, self).__init__()
        # self.network = torchvision.models.resnet18(pretrained=False)
        scale = 1
        self.network = ResNet_cifar(
                        dataset="cifar10",
                        resnet_size=20,
                        scaling=scale,
                        group_norm_num_groups=2,
                        freeze_bn=True,
                        freeze_bn_affine=True,
                    )
        self.n_outputs = 64
        self.n_outputs_feature = 16 * 4 * 4
        self.n_outputs_feature = 32 * 2 * 2
        self.n_outputs_feature = 64

        # self.network = remove_batch_norm_from_resnet(self.network)
        nc = input_shape[0]
        # self.feature = self.network.feature


        # self.freeze_bn()
        self.hparams = hparams
        # self.dropout = nn.Dropout()

    def forward(self, x):
        """Encode x into a feature vector of size n_outputs."""
        # print(x.shape)
        return self.network(x)

    def train(self, mode=True):
        """
        Override the default train() to freeze the BN parameters
        """
        super().train(mode)
        # self.freeze_bn()

    def freeze_bn(self):
        for m in self.network.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

class CIFAR_Vgg(nn.Module):

    def __init__(self, input_shape):
        super(CIFAR_Vgg, self).__init__()
        self.network = torchvision.models.vgg11(pretrained=False).features
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.n_outputs = 512

    def forward(self, x):

        x = self.network(x)
        x = self.avgpool(x)
        return x.squeeze()

class CIFAR_resnet(nn.Module):

    def __init__(self, input_shape):
        super(CIFAR_resnet, self).__init__()
        self.network = torchvision.models.resnet18(pretrained=False)
        self.network.fc = Identity()
        self.n_outputs = 512

    def forward(self, x):

        x = self.network(x)
        # x = self.avgpool(x)
        return x.squeeze()






class MNIST_CNN(nn.Module):
    """
    Hand-tuned architecture for MNIST.
    Weirdness I've noticed so far with this architecture:
    - adding a linear layer after the mean-pool in features hurts
        RotatedMNIST-100 generalization severely.
    """
    n_outputs = 128

    def __init__(self, input_shape):
        super(MNIST_CNN, self).__init__()
        self.conv1 = nn.Conv2d(input_shape[0], 64, 3, 1, padding=1)
        self.conv2 = nn.Conv2d(64, 128, 3, stride=2, padding=1)
        self.conv3 = nn.Conv2d(128, 128, 3, 1, padding=1)
        self.conv4 = nn.Conv2d(128, 128, 3, 1, padding=1)
        # self.projection_head = MLP(128, 256, {'mlp_width': 256, 'mlp_dropout': 0, 'mlp_depth': 2})

        self.bn0 = nn.GroupNorm(8, 64)
        self.bn1 = nn.GroupNorm(8, 128)
        self.bn2 = nn.GroupNorm(8, 128)
        self.bn3 = nn.GroupNorm(8, 128)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.bn0(x)

        x = self.conv2(x)
        x = F.relu(x)
        x = self.bn1(x)

        x = self.conv3(x)
        x = F.relu(x)
        x = self.bn2(x)

        x = self.conv4(x)
        x = F.relu(x)
        x = self.bn3(x)

        x = self.avgpool(x)
        x = x.view(len(x), -1)
        return x
        # return self.projection_head(x)


class ContextNet(nn.Module):
    def __init__(self, input_shape):
        super(ContextNet, self).__init__()

        # Keep same dimensions
        padding = (5 - 1) // 2
        self.context_net = nn.Sequential(
            nn.Conv2d(input_shape[0], 64, 5, padding=padding),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.Conv2d(64, 64, 5, padding=padding),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.Conv2d(64, 1, 5, padding=padding),
        )

    def forward(self, x):
        return self.context_net(x)


def Featurizer(input_shape, hparams):
    """Auto-select an appropriate featurizer for the given input shape."""
    if len(input_shape) == 1:
        return MLP(input_shape[0], hparams["mlp_width"], hparams)
    elif input_shape[1:3] == (28, 28):
        return MNIST_CNN(input_shape)
    elif input_shape[1:3] == (32, 32):
        # print('here')
        # return ResNet(input_shape, hparams)
        # return CIFAR_resnet(input_shape)
        # return ViT(input_shape, hparams)
        return CIFAR_Vgg(input_shape)
        # return MNIST_CNN(input_shape)
        # return wide_resnet.Wide_ResNet(input_shape, 16, 2, 0.)
    elif input_shape[1:3] == (224, 224):
        # return ResNet(input_shape, hparams)
        return CIFAR_Vgg(input_shape)
    else:
        raise NotImplementedError


def Classifier(in_features, out_features, is_nonlinear=False):
    # return cct_7_3x1_32_c100().classifier
    if is_nonlinear:
        return torch.nn.Sequential(
            torch.nn.Linear(in_features, in_features // 2),
            torch.nn.ReLU(),
            torch.nn.Linear(in_features // 2, in_features // 4),
            torch.nn.ReLU(),
            torch.nn.Linear(in_features // 4, out_features))
    else:
        return torch.nn.Linear(in_features, out_features)


class WholeFish(nn.Module):
    def __init__(self, input_shape, num_classes, hparams, weights=None):
        super(WholeFish, self).__init__()
        featurizer = Featurizer(input_shape, hparams)
        classifier = Classifier(
            featurizer.n_outputs,
            num_classes,
            hparams['nonlinear_classifier'])
        self.net = nn.Sequential(
            featurizer, classifier
        )
        if weights is not None:
            self.load_state_dict(copy.deepcopy(weights))

    def reset_weights(self, weights):
        self.load_state_dict(copy.deepcopy(weights))

    def forward(self, x):
        return self.net(x)
